__doc__ = r"""
.. _MetaNetCDF
:mod:`MetaNetCDF` -- PseudoNetCDF manipulation utilities
========================================================

.. module:: MetaNetCDF
   :platform: Unix, Windows
   :synopsis: Provides tools for manipulating NetCDF-like files
              and variables.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

__all__ = ['add_derived', 'time_avg_new_unit', 'window',
           'newresolution', 'MetaNetCDF', 'WindowFromFile', 'file_master']

from numpy import array, logical_or, repeat, sum
# This Package modules
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariable
from PseudoNetCDF.sci_var import PseudoNetCDFVariables, PseudoNetCDFDimension
from PseudoNetCDF.sci_var import PseudoNetCDFVariableConvertUnit
from PseudoNetCDF.ArrayTransforms import CenterTime


class add_derived(PseudoNetCDFFile):
    """add_derived provides a simple interface to add derived variables
    to a PseudoNetCDFFile interface

    create a new class with the following modifications:
    overwrite __childclass__ with the base class
    overwrite __addvars__ with a list of keys for variable names you
                          intend to derive
    for each key name, create a key interface funciton (e.g. key = DEPTH,
    interface = __DEPTH__)
    """
    __childclass__ = None
    __addvars__ = {}

    def __init__(* args, **kwds):
        self = args[0]
        self.__child = self.__childclass__(* args[1:], **kwds)
        self.dimensions = self.__child.dimensions
        self.variables = PseudoNetCDFVariables(
            self.__variables__,
            self.__child.variables.keys() + self.__addvars__)

    def __variables__(self, key):
        if key in self.__addvars__:
            return getattr(self, '__' + key + '__')()
        else:
            return self.__child.variables[key]


class time_avg_new_unit(PseudoNetCDFFile):
    """
    This base class provides an interface for converting
    instantaneous data to time averaged.  It also provides
    an optional unit conversion.

    ex:
        >>> windfile = wind(wind_path, 65, 83)
        >>> windfile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
        >>> windfile.variables['U'].units
        'm/s'
        >>> windfile.variables['V'].units
        'm/s'
        >>> class windavgnewunit(time_avg_new_unit):
        ...     __reader__ = wind
        >>> windavgfile = windavgnewunit(wind_path, rows, cols,\
                          {'U': 'km/h', 'V': 'mph'})
        >>> windavgfile.dimensions
        {'TSTEP': 24, 'LAY': 28, 'ROW': 65, 'COL': 83}
        >>> windfile.variables['U'].units
        'km/h'
        >>> windfile.variables['V'].units
        'mph'
    """
    __reader__ = None

    def __init__(self, rffile, rows, cols, outunit={}, endhour=True):
        self.__file = self.__reader__(rffile, rows, cols)
        self.createDimension('TSTEP', len(self.__file.dimensions['TSTEP'] - 1))
        self.createDimension('LAY', len(self.__file.dimensions['LAY']))
        self.createDimension('ROW', len(self.__file.dimensions['ROW']))
        self.createDimension('COL', len(self.__file.dimensions['COL']))
        self.createDimension('VAR', len(self.__file.dimensions['VAR']))
        if self.__file.dimensions.has('DATE-TIME'):
            self.createDimension(
                'DATE-TIME', len(self.__file.dimensions['DATE-TIME']))
        else:
            self.createDimension('DATE-TIME', 2)
        self.__outunit = outunit
        self.variables = PseudoNetCDFVariables(
            self.__variables, self.__file.variables.keys())
        self.__timeslice = {True: slice(
            1, None), False: slice(None, -1)}[endhour]
        v = self.createVariable(
            'TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME'), keep=True)
        v[:] = self.__file.variables['TFLAG'][self.__timeslice]
        v.long_name = 'Time flag'
        v.units = 'DATE-TIME'

    def __variables(self, k):
        outunit = self.__outunit.get(k, None)
        var = self.__file.variables[k]
        if outunit is None:
            outunit = var.units
        tmpvar = PseudoNetCDFVariable(self, k, var.typecode(), var.dimensions,
                                      values=CenterTime(var))

        return PseudoNetCDFVariableConvertUnit(self.__decorator(var, tmpvar),
                                               outunit)

    def __decorator(self, ovar, nvar):
        for a, v in ovar.__dict__.items():
            setattr(nvar, a, v)
        return nvar


class window(PseudoNetCDFFile):
    """
    Window takes a netcdf or PseudoNetCDF file and creates
    a windowed version.

    ex:
        >>> windfile = wind(wind_path, 65, 83)
        >>> subsetfile = window(windfile, tslice = slice(8, 18), \
                                          kslice = slice(0, 1), \
                                          jslice = slice(19, 37), \
                                          islice = slice(19, 37))
        >>> windfile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
        >>> windfile.variables['U'].shape
        (25, 28, 65, 83)
        >>> subsetfile.dimensions
        {'TSTEP': 10, 'LAY': 1, 'ROW': 18, 'COL': 18}
        >>> windfile.variables['U'].shape
        (10, 1, 18, 18)

    """

    def __init__(self, ncffile, tslice=slice(None), kslice=slice(None),
                 jslice=slice(None), islice=slice(None)):
        self.__file = ncffile
        self.__idx = [tslice, kslice, jslice, islice]
        self.dimensions = self.__file.dimensions.copy()
        any_non_time_key = [
            k for k in self.__file.variables.keys() if 'TFLAG' not in k][0]
        tmpvar = self.__file.variables[any_non_time_key][self.__idx]
        ntimes, nlays, nrows, ncols = tmpvar.shape
        self.dimensions['TSTEP'] = PseudoNetCDFDimension(None, None, ntimes)
        self.dimensions['LAY'] = PseudoNetCDFDimension(None, None, nlays)
        self.dimensions['ROW'] = PseudoNetCDFDimension(None, None, nrows)
        self.dimensions['COL'] = PseudoNetCDFDimension(None, None, ncols)
        self.variables = PseudoNetCDFVariables(
            self.__variables, self.__file.variables.keys())

    def __variables(self, k):
        from PseudoNetCDF.sci_var import Pseudo2NetCDF
        if 'TFLAG' in k:
            return self.__file.variables[k]

        ov = self.__file.variables[k]
        nv = ov[self.__idx]
        Pseudo2NetCDF().addVariableProperties(nv, ov)
        return nv


class newresolution(PseudoNetCDFFile):
    """
    newresolution converts a netcdf or PseudoNetCDF file
    to a new resolution based on user input.  This class
    updates dimensions and variables, but metadata is unaffected.

    ex:
        >>> kvfile = kv(kv_path, 65, 83)
        >>> newresfile = newresolution(kvfile, (2, 3), 4000, 1000)
        >>> kvfile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
        >>> kvfile.variables['KV'].shape
        (25, 28, 65, 83)
        >>> newresfile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 18, 'COL': 18}
        >>> newresfile.variables['KV'].shape
        (25, 28, 260, 332)
    """

    def __init__(self, ncffile, dimension, oldres, newres,
                 repeat_method=repeat, condense_method=sum, nthick=0):
        from PseudoNetCDF.sci_var import Pseudo2NetCDF
        PseudoNetCDFFile.__init__(self)
        self.__dimension = array(dimension, ndmin=1)
        oldres = array(oldres, ndmin=1)
        newres = array(newres, ndmin=1)
        self.__mesh = newres / oldres.astype('f')
        self.__condense = condense_method
        self.__repeat = repeat_method
        self.__file = ncffile
        self.__nthick = nthick

        if not logical_or((self.__mesh % 1) == 0,
                          (1. / self.__mesh) % 1 == 0).any():
            raise ValueError("One resolution must be a factor of the other.")

        Pseudo2NetCDF().addDimensions(self.__file, self)
        any_non_time_key = [
            k for k in self.__file.variables.keys() if 'TFLAG' not in k][0]
        for dk, dfactor in zip(self.__dimension, 1. / self.__mesh):
            dimo = self.dimensions[dk]
            ndimo = self.createDimension(str(dk), len(dimo) * dfactor)
            ndimo.setunlimited(dimo.isunlimited())
        v = self.__file.variables[any_non_time_key]
        v = self.__method(v)

        self.variables = PseudoNetCDFVariables(
            self.__variables, self.__file.variables.keys())

    def __method(self, a):
        dimensions = self.__dimension
        meshs = self.__mesh
        for dimension, mesh in zip(dimensions, meshs):
            axisi = list(a.dimensions).index(dimension)
            newshape = list(a.shape)
            if mesh < 1:
                def method(a):
                    return self.__repeat(a, (1. / mesh), axisi)
                newshape[axisi] = newshape[axisi] * mesh
            else:
                newshape[axisi:axisi +
                         1] = int(newshape[axisi] // mesh), int(mesh)

                def method(a):
                    return self.__condense(
                        a.reshape(newshape), axisi + 1)
            a = method(a)

        return a

    def __variables(self, k):
        from PseudoNetCDF.sci_var import Pseudo2NetCDF
        if 'TFLAG' in k and (self.__axis != 0).any():
            raise KeyError("Tflag is off limits")
        else:
            ov = self.__file.variables[k]
            v = self.__method(ov)
            Pseudo2NetCDF().addVariableProperties(ov, v)
            return v


class MetaNetCDF(PseudoNetCDFFile):
    """
    MetaNetCDF provides a basic interface for combining files
    and adding derived variables.

    ex:
        >>> kvfile = kv(kv_path, 65, 83)
        >>> zpfile = height_pressure(zp_path, 65, 83)
        >>> windfile = wind(wind_path, 65, 83)
        >>> metfile = MetaNetCDF([kvfile, zpfile, windfile])
        >>> wind_speed_calc = lambda self: (self.variables['U'][:]**2 + \
                                            self.variables['V'][:]**2)**.5
        >>> metfile.addMetaVariable('WS', wind_speed_calc)
        >>> metfile.variables['WS'].shape
        (25, 28, 65, 83)
    """
    __metavars__ = {}

    def addMetaVariable(self, key, func):
        self.variables.addkey(key)
        self.__metavars__[key] = func

    def __init__(self, files):
        self.__files = files
        keys = []
        for f in self.__files:
            for k, d in f.dimensions.items():
                if len(d) == 1 and k == 'LAY':
                    k = 'SURFLAY'
                if k not in self.dimensions.keys():
                    self.createDimension(k, len(d))
            keys.extend(f.variables.keys())
            for k in f.__dict__.keys():
                if k not in self.__dict__.keys():
                    setattr(self, k, getattr(f, k))
        keys = list(set(keys))
        self.variables = PseudoNetCDFVariables(self.__variables, keys)
        self.createDimension('VAR', len(keys) - 1)

    def __getattribute__(self, k):
        try:
            return PseudoNetCDFFile.__getattribute__(self, k)
        except AttributeError:
            for f in self.__files:
                try:
                    return getattr(f, k)
                except Exception:
                    pass
            raise AttributeError("%s not found" % k)

    def childvariables(self, k):
        for f in self.__files:
            if k in f.variables.keys():
                v = f.variables[k]
                if k == 'TFLAG':
                    nvars = len(self.dimensions['VAR'])
                    vals = v[:][:, [0], :].repeat(nvars, 1)
                    v = PseudoNetCDFVariable(self, 'TFLAG', 'i', v.dimensions,
                                             values=vals)
                    v.long_name = 'TFLAG'.ljust(16)
                    v.var_desc = 'TFLAG'.ljust(16)
                    v.units = 'DATE-TIME'

                if k == 'LAY' and k in self.dimensions and len(k.shape) > 1:
                    if v.shape[1] == 1:
                        dims = list(v.dimensions)
                        dims[1] = 'SURFLAY'
                        v.dimensions = tuple(dims)
                return v

    def __variables(self, k):
        if k in self.__metavars__.keys():
            return self.__metavars__[k](self)
        else:
            return self.childvariables(k)
        raise KeyError('%s not in any files' % k)


file_master = MetaNetCDF


def WindowFromFile(WindowThis, WindowFrom):
    """
    WindowFromFile creates a bounding box to window one file
    from the meta data of another file.

    ex:
        >>> wind12k = wind(wind12k_path, 89, 89)
        >>> wind04k = wind(wind04k_path, 65, 83)
        >>> window4k_from_wind12k = WindowFromFile(wind12k, wind04k)
        >>> V12k = wind12k.variables['V']
        >>> V04k = wind04k.variables['V']
        >>> V04k_12k = window4k_from_wind12k.variables['V']
        >>> V04k_12k.shape
        (25, 28, 89, 89)
        >>> V04k_12k.shape
        (25, 28, 65, 83)
        >>> V04k_12k.shape # Note the trimmed buffer cells
        (25, 28, 21, 27)
    """
    xoffset = abs(WindowThis.XORIG - WindowFrom.XORIG)
    ioffset = 0
    while xoffset != 0 and not xoffset % WindowThis.XCELL:
        xoffset += WindowFrom.XCELL
        ioffset += 1
    istart = xoffset / WindowThis.XCELL

    yoffset = abs(WindowThis.YORIG - WindowFrom.YORIG)
    joffset = 0
    while yoffset != 0 and not yoffset % WindowThis.YCELL:
        yoffset += WindowFrom.YCELL
        joffset += 1
    jstart = yoffset / WindowThis.YCELL

    iend = istart + int((WindowFrom.NCOLS - 2 * ioffset) *
                        WindowFrom.XCELL / WindowThis.XCELL)
    jend = jstart + int((WindowFrom.NROWS - 2 * joffset) *
                        WindowFrom.YCELL / WindowThis.YCELL)
    islice = slice(istart, iend)
    jslice = slice(jstart, jend)

    outf = MetaNetCDF([WindowThis])

    def AddMetaVar(k, jslice, islice):
        outf.addMetaVariable(k, lambda self: self.childvariables(k)[
                             :, :, jslice, islice])
    for k in WindowThis.variables.keys():
        AddMetaVar(k, jslice, islice)
    return outf



from __future__ import print_function, unicode_literals
import os
import sys
from warnings import warn
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from ._getreader import getreaderdict, pncopen
from ._getwriter import getwriterdict
from .conventions.ioapi import add_ioapi_from_cf
from .conventions.ioapi import add_cf_from_wrfioapi, add_cf_from_ioapi
from PseudoNetCDF import PseudoNetCDFFile
from PseudoNetCDF.netcdf import NetCDFFile
try:
    from netCDF4 import MFDataset
except ImportError:
    pass


from .sci_var import reduce_dim, mesh_dim, slice_dim, getvarpnc, extract
from .sci_var import mask_vals, seqpncbo, pncexpr, stack_files, add_attr
from .sci_var import convolve_dim, manglenames, removesingleton, merge
from .sci_var import extract_from_file, pncrename, WrapPNC


from argparse import SUPPRESS


conv2conv = {}
conv2conv[('ioapi', 'cf')] = add_cf_from_ioapi
conv2conv[('wrfioapi', 'cf')] = add_cf_from_wrfioapi
conv2conv[('cf', 'ioapi')] = add_ioapi_from_cf

allreaders = getreaderdict()
allwriters = getwriterdict()
globals().update(allreaders)
_readernames = [(k.count('.') if k[:1] != '_' else 9999, k)
                for k in getreaderdict().keys()]
_readernames.sort()
_readernames = [k for c, k in _readernames]
_writernames = [(k.count('.'), k) for k in getwriterdict().keys()]
_writernames.sort()
_writernames = [k for c, k in _writernames]
_coordkeys = ("time time_bounds TFLAG ETFLAG latitude latitude_bounds " +
              "longitude longitude_bounds lat lat_bnds lon lon_bnds " +
              "etam_pressure etai_pressure layer_bounds layer47 layer").split()


class PNCArgumentParser(ArgumentParser):
    def exit(self, status=0, message=None):
        import sys as _sys
        if message:
            self._print_message(message, _sys.stderr)


class _HelpAction(Action):
    def __init__(self,
                 option_strings,
                 dest=SUPPRESS,
                 default=SUPPRESS,
                 help=None):
        super(_HelpAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        parser.print_help()
        setattr(namespace, self.dest, True)


class _HelpListFormats(Action):
    def __init__(self,
                 option_strings,
                 dest=SUPPRESS,
                 default=SUPPRESS,
                 help=None):
        super(_HelpListFormats, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        print('The formats listed below are available ' +
              'for the following options')
        print('-f FORMAT or --format FORMAT or --help-format FORMAT')
        print('where FORMAT is one of the options below')
        print('\t' + '\n\t'.join(_readernames))
        print('**Some readers are listed twice (e.g., without dotted form)')
        setattr(namespace, 'help', True)


class _HelpFormat(Action):
    def __init__(self,
                 option_strings,
                 dest=SUPPRESS,
                 default=SUPPRESS,
                 help=None):
        super(_HelpFormat, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=1,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        file_format = values[0].split(',')[0]
        print('')
        print('All formats require a "path". Some formats also ')
        print('take extra arguments. All arguments other than ')
        print('the input path must be specified using keyword')
        print('arguments.')
        print('')
        helpformat = allreaders[file_format]
        try:
            import inspect
            print('Example:')
            idef = inspect.getargspec(helpformat.__init__)
            args = idef.args[2:]
            if idef.defaults is None:
                defs = ['<VAL>'] * len(args)
            else:
                defs = (len(args) - len(idef.defaults)) * \
                    ['<VAL>'] + list(idef.defaults)
            longform = (
                'pncdump -f ' +
                ','.join([file_format] +
                         [karg + "=%s" % kdef
                          for karg, kdef in zip(args, defs)]) + ' path')
            shortform = (
                'pncdump -f ' +
                ','.join([file_format] +
                         [karg + "=%s" % kdef
                          for karg, kdef in zip(args, defs)
                          if kdef == '<VAL>']) +
                ' path')
            print(shortform)
            if longform != shortform:
                print('')
                print('Extended example with keywords:')
                print(longform)
                print('')
                print('* Any keyword with a non "<VAL>" default ' +
                      'can be omitted.')
            print('')
        except Exception:
            print('** Short help not available for ' + file_format)
            print('')

        if 'y' == input('Hit Y/y to see detailed help\n').lower():
            help(helpformat)
        setattr(namespace, self.dest, values)
        setattr(namespace, 'help', True)


class AggCommaString(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values.split(','))


_plotcmds = ['plot', 'plot2d', 'plotts', 'plotprofile', 'plotscatter']
_allcmds = ['gen', 'dump', 'eval', 'map'] + _plotcmds


def add_basic_options(parser):
    padd = parser.add_argument
    parser.add_argument('ifiles', nargs='*',
                        help='path to a file formatted as type -f')
    parser.add_argument("--verbose", dest="verbose", action="count",
                        default=0, help="Provides verbosity with pncgen")

    try:
        padd("--help",
             dest="help", action=_HelpAction, default=False,
             help="Displays help")
    except Exception:
        pass

    padd('--pnc', action='append', default=[],
         help='Set of pseudonetcdf commands to be process separately')

    padd("-f", "--format", dest="format", default='netcdf',
         metavar='{see --list-formats for choices}',
         help=("File format (default netcdf), can be one of the choices " +
               "listed, or an expression that evaluates to a reader. " +
               "Keyword arguments are passed via ,kwd=value."))

    padd("--list-formats", dest="help", default=None,
         action=_HelpListFormats, help="Show format options for -f")

    padd("--help-format", dest='helpformat',
         action=_HelpFormat, default=None,
         help="Show help for file format (must be one of the options for -f)")

    padd("--sep", dest="separator", action='store_true',
         default=False,
         help=("Used to separate groups of arguments for parsing (e.g., " +
               "pncgen -- [options1] file(s)1 [--sep [options2] file(s)2 " +
               "[... [--sep [optionsN] file(s)N]] "))

    padd("--inherit", dest="inherit", action="store_true",
         default=False,
         help=("Allow subparsed sections (separated with -- and --sep) to " +
               "inherit from global options (-f, --format is always " +
               "inherited)."))

    padd("--diskless", dest="diskless", action="store_true",
         default=False,
         help="Load file into memory; useful for subsequent processing")

    padd("--mangle", dest="mangle", action="store_true",
         default=False, help="Remove non-standard ascii from names")

    padd("--rename", dest="rename", action="append", default=[],
         help=("Provide pairs of strings to be substituted " +
               "--rename=type,oldkey,newkey (type: v = variable; d = " +
               "dimension;)"))

    padd("--remove-singleton", dest="removesingleton",
         type=lambda x: [k for k in x.split(',') if k != ''], default=[],
         help="Remove singleton (length 1) dimensions")

    padd("--coordkeys", dest="coordkeys", type=str,
         action=AggCommaString, default=_coordkeys, metavar="key1,key2",
         help="Variables to be ignored in pncbo.")

    padd("-v", "--variables", dest="variables", default=None,
         action=AggCommaString, metavar='varname1[,varname2[,...,varnameN]',
         help=("Variable names or regular expressions (using match) " +
               "separated by ','. If a group(s) has been specified, only " +
               "variables in that (those) group(s) will be selected."))

    padd("-a", "--attribute", dest="attribute", type=str,
         action="append", default=[],
         metavar="att_nm,var_nm,mode,att_typ,att_val",
         help=("Variables have attributes that can be added following nco " +
               "syntax (--attribute att_nm,var_nm,mode,att_typ,att_val); " +
               "mode = a,c,d,m,o and att_typ = f,d,l,s,c,b; att_typ is any " +
               "valid numpy type."))

    padd("-m", "--mask", dest="masks", action="append",
         default=[],
         help=("Masks to apply (e.g., greater,0 or less,0 or values,0, or " +
               "where,(time[:]%%24<12)[:,None].repeat(10,1))"))

    padd("--from-convention", dest="fromconv", type=str,
         default=None, help="From convention currently only support ioapi")

    padd("--to-convention", dest="toconv", type=str,
         default="cf", help="To convention currently only supports cf")

    padd("--stack", dest="stack", type=str,
         help="Concatentate (stack) files on the dimension.")

    padd("--merge", dest="merge", action='store_true',
         help="Combine variables into one file")

    padd("-s", "--slice", dest="slice", type=str,
         action="append", default=[], metavar='dim,start[,stop[,step]]',
         help=("Variables have dimensions (time, layer, lat, lon), which " +
               "can be subset using dim,start,stop,stride (e.g., " +
               "--slice=layer,0,47,5 would sample every fifth layer " +
               "starting at 0)"))

    padd("-r", "--reduce", dest="reduce", type=str,
         action="append", default=[], metavar='dim,function[,weight]',
         help=("Variable dimensions can be reduced using dim,function," +
               "weight syntax (e.g., --reduce=layer,mean,weight). " +
               "Weighting is not fully functional."))

    padd("--mesh", dest="mesh", type=str, action="append",
         default=[], metavar='dim,weight,function',
         help=("Variable dimensions can be meshed using dim,function," +
               "weight syntax (e.g., --mesh=time,0.5,mean)."))

    padd("-c", "--convolve", dest="convolve", type=str,
         action="append", default=[], metavar='dim,mode,wgt1,wgt2,...wgtN',
         help=("Variable dimension is reduced by convolve function " +
               "(dim,mode,wgt1,wgt2,...wgtN)"))

    padd("-e", "--extract", dest="extract", action="append",
         default=[],
         help=("lon/lat coordinates to extract lon1,lat1/lon2,lat2/lon3," +
               "lat3/.../lonN,latN"))

    padd("--extract-file", dest="extractfile", action="append",
         default=[], help="pncparse options for file")

    padd("--extractmethod", dest="extractmethod", type=str,
         default='nn', help="Method for extraction",
         choices=['ll2ij', 'nn', 'linear', 'cubic', 'quintic', 'KDTree'])

    padd("--op-typ", dest="operators", type=str,
         action='append', default=[],
         help=("Operator for binary file operations. Binary file operations " +
               "use the first two files, then the result and the next file, " +
               "etc. Use " +
               " or ".join(['//', '<=', '%%', 'is not', '>>', '&', '==', '!=',
                            '+', '*', '-', '/', '<', '>=', '**', '>', '<<',
                            '|', 'is', '^'])))

    padd("--expr", dest="expressions", type=str,
         action='append', default=[],
         help="Generic expressions to execute in the context of the file.")

    padd("--exprscript", dest="expressionscripts", type=str,
         action='append', default=[],
         help="Generic expressions to execute in the context of the file.")


def add_2d_options(parser):
    parser.add_argument('--swapaxes', action='store_true',
                        help='Swap x-y axes')


def add_eval_options(parser):
    padd = parser.add_argument
    from .pnceval import __all__ as evalfuncs
    add_interactive_options(parser)
    padd('--funcs', default=evalfuncs,
         type=lambda x: x.split(','),
         help=('Functions to evaluate split by , (default: %s)' %
               ','.join(evalfuncs)))


def add_map_options(parser):
    padd = parser.add_argument
    padd("--iter", dest="iter", action='append', default=[],
         help=("Create plots for each element of specified dimension " +
               "(e.g., --iter=time)."))

    padd("--resolution", dest="resolution",
         choices=['c', 'i', 'h'], default='c',
         help=("Use coarse (c), intermediate (i), or high (h) " +
               "resolution for maps."))

    padd("--no-coastlines", dest="coastlines",
         action='store_false',
         help="Disable coastlines by setting equal to False")

    padd("--no-countries", dest="countries",
         action='store_false',
         help="Disable countries by setting equal to False")

    padd("--states", dest="states", action='store_true',
         help="Enable states by setting equal to True")

    padd("--counties", dest="counties", action='store_true',
         help="Enable counties by setting equal to True")

    padd("--shapefile", "--shapefiles", dest="shapefiles",
         type=str, action='append', default=[],
         help=("Enable custom shapefiles (must be lon, lat). Keyword " +
               "arguments for display can be added with commas (see -f " +
               "for syntax)."))


def add_dump_options(parser):
    padd = parser.add_argument
    parser.add_argument("-H", "--header", dest="header",
                        action="store_true", default=False)

    parser.add_argument("-t", "--timestring", dest="timestring",
                        action="store_true", default=False)

    padd("--full-indices", dest="full_indices", default=None,
         metavar="[c|f]", choices=['c', 'f'],
         help=("Provide indices in CDL using either C or Fortran style " +
               "indexes. C style is 0-based and ordered from slowest " +
               "iterating dimension to fastest. Fortran style is 1-based " +
               "and ordered from fastest to slowest iterating dimension"))

    padd("-l", "--length", dest="line_length", type=int,
         default=80, metavar="LEN",
         help="CDL line length (pncdump only)")

    padd("--float-precision", dest="float_precision", type=int,
         default=8, metavar="FDIG",
         help="single precision digitis (default 8; pncdump only)")

    padd("--double-precision", dest="double_precision",
         type=int, default=16, metavar="PDIG",
         help="pdig double precision digits (default 16; pncdump only)")

    padd("--dump-name", dest="cdlname", type=str,
         default=None, help="Name for display in ncdump")


def add_output_options(parser):
    padd = parser.add_argument
    padd("-O", "--clobber", dest="clobber",
         action='store_true', default=False,
         help="Overwrite existing file if necessary.")

    ncex = "NETCDF3_CLASSIC, NETCDF4_CLASSIC, NETCDF4"
    padd("--out-format", dest="outformat",
         default="NETCDF4_CLASSIC", type=str,
         help="File output format (e.g., " + ncex + ";pncgen only)",
         choices=ncex.split(', ') + _writernames)

    padd("--mode", dest="mode", type=str, default="w",
         choices='w a r+ ws as r+s'.split(),
         help=("File mode for writing (w, a or r+ or with unbuffered " +
               "writes ws, as, or r+s; pncgen only)."))

    padd('outpath', default=None, type=str,
         help='path to a output file formatted as --out-format')


def add_interactive_options(parser):
    try:
        parser.add_argument("-i", "--interactive", dest="interactive",
                            action='store_true', default=False,
                            help="Use interactive mode")
    except Exception:
        pass


def add_plot_options(parser):
    padd = parser.add_argument
    add_interactive_options(parser)
    padd("--matplotlibrc", dest="matplotlibrc",
         type=lambda x: x.split(';'), action='append', default=[],
         help='rc options for matplotlib')
    padd("--figure-keywords", dest='figure_keywords',
         type=lambda x: eval('dict(' + x + ')'), default=dict(),
         help='options for figure')

    padd("--axes-keywords", dest='axes_keywords',
         type=lambda x: eval('dict(' + x + ')'),
         default=dict(), help='options for axes')

    padd("--plot-commands", dest="plotcommands", type=str,
         action='append', default=[],
         help=("Plotting functions to call for all variables expressions " +
               "to execute in the context of the file. The figure object " +
               "will always be 'fig'."))

    padd("--figformat", dest="figformat", type=str,
         default='png', help="Any format supported by matplotlib")

    padd("--norm", dest="normalize", type=str, default=None,
         help=("Typical examples Normalize(), LogNorm(), BoundaryNorm([0, " +
               "10, 20, 30, 40], ncolors = 256)"))

    padd("--colorbar-formatter", dest="colorbarformatter",
         type=str, default=None,
         help=("Typical examples LogFormatter(labelOnlyBase = False), " +
               "ScalarFormatter(), '%%3g'"))

    padd("--overlay", dest="overlay", action='store_true',
         help="Enable overlay by setting equal to True")

    padd('--no-squeeze', dest='squeeze', default=True,
         action='store_false',
         help=('Squeeze automatically removes singleton dimensions; ' +
               'disabling requires user to remove singleton dimensions ' +
               'with --remove-singleton option'))

    padd("--figroot", help='Path for base of figure (suffixed by action)')


_pncepilog = """
Detailed Steps
--------------

PseudoNetCDF has many operations and the order often matters. The order is
consistent with the order of options in the formatted help. The default order
is summarized as:

1. Open with specified reader (-f)
2. Select subset of variables (-v)
2. Add attributes (-a)
4. Apply masks (--mask)
5. Add conventions to support later operations (--to-conv, --from-conv)
6. Combine files via stacking on dimensions (--stack)
7. Slice dimensions (-s --slice)
8. Reduce dimensions (-r --reduce)
9. Convolve dimensions (-c)
10. Extract specific coordinates (--extract)
11. Remove singleton dimensions (--remove-singleton)
12. Apply expressions (--expr then --exprscripts)
13. Apply binary operators (--op_typ)

To impose your own order, use standard options (global options) and then
use -- to force positional interpretation of remaining options. In remaining
options, use --sep to separate groups of files and options to be evaluated
before any global operations."""


def getparser2(actions=False):
    rdesc = RawDescriptionHelpFormatter
    parent_parser = PNCArgumentParser(description="PseudoNetCDF",
                                      formatter_class=rdesc,
                                      add_help=False,
                                      prog='PNC')
    if actions:
        add_action_commands(
            parent_parser, withbasic=True, add_help=False)
    else:
        add_basic_options(parent_parser)
    return parent_parser


def getparser(has_ofile, plot_options=False, map_options=False,
              interactive=False, actions=False):
    """getparser produces a parser for PseudoNetCDF

Parameters
----------
has_ofile : boolean
            Requires the outpath option and processes existence check
plot_options : boolean, optional
               Adds options for plotting including matplotlibrc, spatial
               overlays, and normalization options, default is False
interactive : boolean, optional
              Adds for interactive option, default is False

Returns
-------
parser : ArgumentParser with options that are processable with pncparse

    """
    parser = PNCArgumentParser(description="""PseudoNetCDF Argument Parsing


""", formatter_class=RawDescriptionHelpFormatter)
    parser.epilog = _pncepilog
    add_basic_options(parser)
    if interactive:
        add_interactive_options(parser)

    if actions:
        add_action_commands(parser, add_help=False)
    else:
        # Only has output file if pncgen is called.
        if has_ofile:
            add_output_options(parser)
        else:
            add_dump_options(parser)

        if plot_options:
            add_plot_options(parser)
        if map_options:
            add_map_options(parser)

    return parser


def pncparse(has_ofile=False, plot_options=False, map_options=False,
             interactive=False, args=None, parser=None, ifiles=[]):
    """
Parameters
----------
has_ofile : boolean, optional
            Requires the outpath option and processes existence check
            default is False
plot_options : boolean, optional
               Processes matplotlib options before loading matplotlib
               (preprocessing important for backend), default is False.
interactive : boolean, optional
              Only relevant if parser is not provided (see getparser),
              default is False.
args : list or string, optional
       args are usually taken from the command-line, but can be provided
       in a function call, default is None.
parser : AgumentParser object, optional
         pncparser parser, default getparser(has_ofile,
                                             plot_options = plot_options,
                                             interactive = interactive)

Returns
-------
(ifiles, args)

ifiles : list of PseudoNetCDFFiles
args : args as parsed

    """
    inputargs = args
    if parser is None:
        parser = getparser(has_ofile, plot_options=plot_options,
                           map_options=map_options, interactive=interactive)
    subparser = getparser(
        has_ofile=False, plot_options=plot_options, interactive=interactive)
    try:
        args = parser.parse_args(args=args)
    except Exception as e:
        raise e
    except BaseException:
        return [], args

    args.inputargs = inputargs or sys.argv
    inpaths = getattr(args, 'ifiles', [])
    inpnc = getattr(args, 'pnc', [])
    inopts = len(inpaths + inpnc)
    if len(ifiles) == 0 and inopts == 0:
        if not getattr(args, 'help', False):
            parser.print_help()

        print('\n*ifile or pnc are required')
        return [], args

    subargs = split_positionals(subparser, args)
    ifiles = [ifile for ifile in ifiles]
    ipaths = ['unknown'] * len(ifiles)
    for subarg in subargs:
        ipaths.extend(subarg.ifiles)
        ifiles.extend(pncprep(subarg)[0])

    # ifile is set to results from parsing
    # this includes the standard ifile
    args.ifiles = ifiles
    args.ipath = ipaths
    # if args.stack is not None:
    #    pass
    # elif (len(args.ifiles) == 0 or
    #       (len(args.ifiles) - len(args.operators) - has_ofile) > 1):
    #    print('Too many arguments', len(args.ifiles),
    #          len(args.operators), has_ofile)
    #    parser.print_help()
    #    return
    if has_ofile or not getattr(args, 'outpath', None) is None:
        if not args.clobber and os.path.exists(args.outpath):
            parser.error(message=(('Output path (%s) exists; enable ' +
                                   'clobber (--clobber or -O) to ' +
                                   'overwrite.') % (args.outpath,)))

    if plot_options or hasattr(args, 'matplotlibrc'):
        from matplotlib import rcParams
        for rcassigns in args.matplotlibrc:
            for rcassign in rcassigns:
                key, value = rcassign.split('=')
                rcParams[key] = value

    out = pncprep(args)
    return out


def add_action_commands(parser, withbasic=False, add_help=True):
    subparsers = parser.add_subparsers(dest='subcommand',
                                       help='Adding commands for subparsing')
    dumpparser = subparsers.add_parser('dump',
                                       description=('Output Common Data ' +
                                                    'Language (like from ' +
                                                    'ncdump)'),
                                       help='sub-command help',
                                       add_help=add_help)
    add_dump_options(dumpparser)
    if withbasic:
        add_basic_options(dumpparser)

    genparser = subparsers.add_parser('gen',
                                      description=('Like ncgen with more ' +
                                                   'format support'),
                                      help='sub-command help',
                                      add_help=add_help)
    if withbasic:
        add_basic_options(genparser)
    add_output_options(genparser)

    evalparser = subparsers.add_parser('eval',
                                       description='Evaluation options.',
                                       help='sub-command help',
                                       add_help=add_help)
    add_eval_options(evalparser)
    if withbasic:
        add_basic_options(evalparser)

    mapparser = subparsers.add_parser('map',
                                      description='Make maps quickly',
                                      help='sub-command help',
                                      add_help=add_help)
    add_plot_options(mapparser)
    add_map_options(mapparser)
    if withbasic:
        add_basic_options(mapparser)

    for plotcmd in _plotcmds:
        plotparser = subparsers.add_parser(
            plotcmd, description='Plot data', add_help=add_help)
        add_plot_options(plotparser)
        if withbasic:
            add_basic_options(plotparser)
        if plotcmd == 'plotprofile':
            from .plotutil import add_vertprofile_options
            add_vertprofile_options(plotparser)


def do_actions(outargs):
    if 'subcommand' not in outargs:
        return
    if outargs.subcommand == 'dump':
        from .pncdump import pncdump
        for ifile in outargs.ifiles:
            pncdump(ifile, header=outargs.header,
                    full_indices=outargs.full_indices,
                    line_length=outargs.line_length,
                    float_precision=outargs.float_precision,
                    double_precision=outargs.double_precision,
                    timestring=outargs.timestring, name=outargs.cdlname)

    elif outargs.subcommand == 'gen':
        from .pncgen import pncgen
        nout = len(outargs.ifiles)
        if nout != 1:
            raise IOError(
                'pncgen can output only 1 file; user requested %d' % nout)
        ifile, = outargs.ifiles
        pncgen(ifile, outargs.outpath, outmode=outargs.mode,
               format=outargs.outformat, verbose=outargs.verbose)
    elif outargs.subcommand == 'eval':
        from .pnceval import pnceval
        nin = len(outargs.ifiles)
        if nin != 2:
            raise IOError(
                'pnceval requires 2 files; user requested %d' % nin)
        pnceval(outargs)
    elif outargs.subcommand == 'map':
        if getattr(outargs, 'outpath', None) is None:
            outargs.outpath = outargs.figroot
        from .plotutil.pncmap import makemaps
        makemaps(outargs)
    elif outargs.subcommand in _plotcmds:
        if getattr(outargs, 'outpath', None) is None:
            outargs.outpath = outargs.figroot
        from . import plotutil
        getattr(plotutil, outargs.subcommand)(outargs)


def PNC(*args, **kwds):
    """
Arguments:
    args - Command Line arguments/options for PseudoNetCDF
           for full list of potential args PNC('--help')
    ifiles - (optional) pre-loaded input files
    actions -  (default: False)
        False: only open files do not make outputs
        True: enable dump,gen,map,etc output actions
              for action options see subparsers help
              e.g., PNC('dump', '--help', actions = True)

Returns:
    out - Namespace object with parsed arguments
          including a list of processed files (out.ifiles)

Example:
    # Single File
    out = PNC('--format=netcdf', inpath)
    infile = out.ifiles[0]
    O3 = infile.variables['O3']

    # Multiple Files
    out = PNC('--format=netcdf', inpath1, inpath2)
    infile1, infile2 = out.ifiles
    O3_1 = infile1.variables['O3']
    O3_2 = infile2.variables['O3']

    # With Actions
    out = PNC('dump', '--variables=O3', '--format=netcdf', inpath)
netcdf icartt/dc3-mrg60-dc8_merge_20120518_R7_thru20120622.nc {
dimensions:
        POINTS = 6817 ;

variables:
        double O3(TSTEP, LAY, ROW, COL);
                O3_ESRL:units = "ppbV" ;
                O3_ESRL:standard_name = "Ozone" ;
                O3_ESRL:missing_value = -999999 ;
    ...
    """
    ifiles = kwds.pop('ifiles', [])
    actions = kwds.pop('actions', None)
    if actions is None:
        actions = args[0] in _allcmds

    parser = getparser2(actions=actions)
    nfiles = len(ifiles)
    ifiles, outargs = pncparse(args=args, ifiles=ifiles, parser=parser, **kwds)
    if nfiles > 0 and len(ifiles) != nfiles:
        raise IOError('PNC can only use files or paths, not both at this time')
    do_actions(outargs)

    return outargs


def pnc(*args, **kwds):
    """
Arguments - see PNC
Returns:
    file(s) - single file or, if more than 1 file is returned a list of files
    """
    ifiles = kwds.pop('ifiles', [])
    actions = kwds.pop('actions', None)
    out = PNC(*args, ifiles=ifiles, actions=actions, **kwds).ifiles
    if len(out) == 1:
        return out[0]
    else:
        return out


def split_positionals(parser, args):
    import shlex
    positionals = args.ifiles
    parser.set_defaults(
        **dict([(k, v) for k, v in args._get_kwargs()
                if args.inherit or k == 'format']))
    ins = [shlex.split(pnc) for pnc in args.pnc]
    last_split = 0
    for i in range(len(positionals)):
        if positionals[i] in ('--', '--sep'):
            ins.append(positionals[last_split:i])
            last_split = i + 1
    ins.append(positionals[last_split:])
    try:
        outs = list(map(parser.parse_args, ins))
        for outns, inargs in zip(outs, ins):
            outns.inputargs = inargs
    except Exception as e:
        print(str(e))
        raise e
    except BaseException as be:
        print(str(be))
        return [], args

    for out in outs:
        if out.cdlname is None:
            out.cdlname = ', '.join(out.ifiles)
    if len(outs) == 1 and getattr(args, 'cdlname', None):
        args.cdlname = outs[0].cdlname
    return outs


def pncprep(args):
    # nifiles = len(args.ifiles) - has_ofile
    ipaths = args.ifiles[:]
    fs = getfiles(ipaths, args)
    fs = subsetfiles(fs, args)
    fs = seqpncbo(args.operators, fs, coordkeys=args.coordkeys)
    for expr in args.expressions:
        fs = [pncexpr(expr, f) for f in fs]
    for script in args.expressionscripts:
        expr = open(script).read()
        fs = [pncexpr(expr, f) for f in fs]
    args.ifiles = fs
    if getattr(args, 'cdlname', None) is None:
        try:
            args.cdlname = args.ipath[0]
        except Exception:
            args.cdlname = 'unknown'
    return fs, args


def subsetfiles(ifiles, args):
    fs = []
    for f in ifiles:
        for opts in args.slice:
            f = slice_dim(f, opts)
        for opts in args.reduce:
            f = reduce_dim(f, opts, metakeys=args.coordkeys)
        for opts in args.mesh:
            f = mesh_dim(f, opts)
        for opts in args.convolve:
            f = convolve_dim(f, opts)
        if len(args.extract) > 0:
            f = extract(f, args.extract, method=args.extractmethod)
        if len(args.extractfile) > 0:
            extractfiles = []
            import shlex
            for extractfile in args.extractfile:
                extractfile, options = pncparse(
                    has_ofile=False, args=shlex.split(extractfile))
                extractfiles.extend(extractfile)

            f = extract_from_file(f, extractfiles, method=args.extractmethod)
        for rd in args.removesingleton:
            f = removesingleton(f, rd)
        fs.append(f)
    return fs


def getfiles(ipaths, args):
    fs = []
    for ipath in ipaths:
        format_options = args.format.split(',')
        file_format = format_options.pop(0)
        format_options = eval('dict(' + ', '.join(format_options) + ')')
        if isinstance(ipath, (PseudoNetCDFFile, NetCDFFile, MFDataset)):
            f = ipath
        elif isinstance(ipath, (str,)):
            try:
                # Most common usecase; open like pncopen
                f = pncopen(ipath, format=file_format, **format_options)
            except Exception:
                try:
                    allreaders = getreaderdict()
                    if file_format in allreaders:
                        # Unknown problem, try reader directly
                        allreaders[file_format](ipath, **format_options)
                    else:
                        # Maybe the format is an expression?
                        allreaders[file_format](ipath, **format_options)
                        f = eval(file_format)(ipath, **format_options)
                except Exception as e:
                    etmp = ('Unable to open path with %s(path, **%s)\n\t' +
                            'path="%s"\n\terror="%s"')
                    oute = IOError(etmp % (file_format, str(format_options),
                                           ipath, str(e)))
                    raise oute  # from e
        else:
            warn('File is type %s, which is unknown' % type(ipath))
            f = ipath

        history = getattr(f, 'history', getattr(f, 'HISTORY', ''))
        history += ' '.join(args.inputargs) + ';'
        laddconv = args.fromconv is not None and args.toconv is not None
        # Need to add backward compatibility
        # lslice = len(args.slice + args.reduce) > 0
        # lexpr = len(args.expressions) > 0
        if args.variables is not None:
            f = getvarpnc(f, args.variables, coordkeys=args.coordkeys)
        elif args.diskless:
            f = getvarpnc(f, None)
        for opts in args.attribute:
            add_attr(f, opts)
        for opts in args.masks:
            f = mask_vals(f, opts, metakeys=args.coordkeys)
        if laddconv:
            if not args.diskless:
                f = WrapPNC(f)
            try:
                conv2conv[args.fromconv, args.toconv](
                    f, coordkeys=args.coordkeys)
            except KeyError as e:
                warn('Cannot add %s from %s; %s' %
                     (args.toconv, args.fromconv, str(e)))

        try:
            setattr(f, 'history', history)
        except Exception:
            pass
        if args.mangle:
            f = manglenames(f)
        for rename in args.rename:
            f = pncrename(f, rename)
        fs.append(f)
    if args.stack is not None:
        fs = [stack_files(fs, args.stack, coordkeys=args.coordkeys)]
    if args.merge:
        fs = [merge(fs)]
    return fs



__all__ = ['arl2time']


def arl2time(YY, MM, DD, HH, mm):
    """
    Convert ARL time pieces to a datetime object

    Parameters
    ----------
    YY : integer year (2 or 4 digits)
    MM : integer month (1-12)
    DD : integer day (1-31)
    HH : integer hour (0-23)
    mm : integer minute (0-59)

    Returns
    -------
    d : datetime object from datetime.strptime
    """
    from datetime import datetime
    strptime = datetime.strptime
    if YY < 100:
        yf = '%y'
    else:
        yf = '%Y'
    out = strptime('%02d-%02d-%02d %02d:%02d+0000' % (YY, MM, DD, HH, mm),
                   yf + '-%m-%d %H:%M%z')
    return out


def arl2timea(year, month, day, hour, minute):
    """
    Convert ARL time piece arrays to a datetime object array

    Parameters (all numpy array
    ----------
    YY : array integer year (2 or 4 digits)
    MM : integer month (1-12)
    DD : integer day (1-31)
    HH : integer hour (0-23)
    mm : integer minute (0-59)

    Returns
    -------
    d : array of datetime object from datetime.strptime
    """
    import numpy as np
    from datetime import datetime
    datei = (year.astype('l') * 100000000 +
             month.astype('l') * 1000000 +
             day.astype('l') * 10000 +
             hour.astype('l') * 100 +
             minute.astype('l'))
    datestrs = np.char.decode(np.char.add(datei.astype('S16'), b'+0000'))
    dates = np.array([datetime.strptime(d, '%y%m%d%H%M%z') for d in datestrs])
    return dates



from __future__ import print_function
from ..core._files import PseudoNetCDFFile
import numpy as np
from datetime import datetime, timedelta


class cmaqomidat(PseudoNetCDFFile):
    def __init__(self, path):
        """
        Arguments
        ---------
        path : str
            path to CMAQ-ready OMI file. This is the output of create_omi
            preprocessor. See notes for general form.

        Notes
        -----
        File has the genearl form

        nlat 17
        nlon 17
        yeardate latitude 180Z 157.5W ...
        1990.203       80  426    415 ...
        ...
        1990.203      -80  260    253 ...
        """
        import pandas as pd
        tmpf = open(path, 'r')
        key1, val1 = tmpf.readline().split()
        key2, val2 = tmpf.readline().split()
        if key1.strip() == 'nlat':
            nlat = eval(val1)
            nlon = eval(val2)

        if key1.strip() == 'nlon':
            nlon = eval(val1)
            nlat = eval(val2)

        self._df = df = pd.read_csv(path, delimiter=r'\s+', skiprows=2)

        latitudes = df.latitude.values.reshape(-1, nlat)
        if not (latitudes[0] == latitudes[:]).all():
            raise ValueError('latitudes should all be in the same order')

        # Latitudes are reordered from top-bottom to bottom-top
        data = self._df.drop(
            columns=['yeardate', 'latitude']
        ).values.reshape(-1, nlat, nlon)[:, ::-1]

        ntime = data.shape[0]

        # Latitudes are reordered from top-bottom to bottom-top
        latc = latitudes[0, ::-1]
        dlat = np.abs(np.diff(latc)).mean() / 2
        latb = np.array([latc - dlat, latc + dlat]).T

        lonc = np.linspace(-180, 180, nlon)
        dlon = np.abs(np.diff(lonc)).mean() / 2
        lonb = np.array([lonc - dlon, lonc + dlon]).T

        self.createDimension('time', ntime)
        self.createDimension('latitude', nlat)
        self.createDimension('longitude', nlon)
        self.createDimension('nv', 2)
        fltyears = np.sort(df.yeardate.unique())
        rdate = datetime(1970, 1, 1)
        dts = []
        for fltyear in fltyears:
            year, fyear = divmod(fltyear, 1)
            year = int(year)
            nextyear = datetime(year + 1, 1, 1)
            thisyear = datetime(year, 1, 1)
            ndays = (nextyear - thisyear).total_seconds() / 3600. / 24.
            date = thisyear + timedelta(days=ndays * fyear)
            dt = (date - rdate).total_seconds() / 3600. / 24.
            dts.append(dt)

        dts = np.array(dts)
        self.createVariable(
            'year', 'd', ('time',),
            units='fractional year'
        )[:] = fltyears
        self.createVariable(
            'time', 'd', ('time',),
            units='days since 1970-01-01 00:00:00+0000',
            values=dts
        )
        self.createVariable(
            'latitude', 'd', ('latitude',),
            units='degrees_north', values=latc
        )
        self.createVariable(
            'latitude_bounds', 'd', ('latitude', 'nv'),
            units='degrees_north', values=latb
        )
        self.createVariable(
            'longitude', 'd', ('latitude',),
            units='degrees_east', values=lonc
        )
        self.createVariable(
            'longitude_bounds', 'd', ('latitude', 'nv'),
            units='degrees_east', values=lonb
        )
        v = self.createVariable(
            'O3', 'd', ('time', 'latitude', 'longitude'),
            missing_value=-1, units='DU'
        )
        v[:] = data
        self.setCoords([
            'time', 'latitude', 'longitude',
            'latitude_bounds', 'longitude_bounds'
        ])



from __future__ import print_function
import os
from collections import OrderedDict
import numpy as np
from PseudoNetCDF.sci_var import PseudoNetCDFVariables
from ._bpch import _general_header_type, bpch1 as oldbpch, bpch_base
import unittest


_datablock_header_type = np.dtype([('bpad1', '>i4'),
                                   ('modelname', 'S20'),
                                   ('modelres', '>f4', 2),
                                   ('halfpolar', '>i4'),
                                   ('center180', '>i4'),
                                   ('epad1', '>i4'),
                                   ('bpad2', '>i4'),
                                   ('category', 'S40'),
                                   ('tracerid', '>i4'),
                                   ('base_units', 'S40'),
                                   ('tau0', '>f8'),
                                   ('tau1', '>f8'),
                                   ('reserved', 'S40'),
                                   ('dim', '>i4', 3),
                                   ('start', '>i4', 3),
                                   ('skip', '>i4'),
                                   ('epad2', '>i4')])
_hdr_size = _datablock_header_type.itemsize


def add_vert(self, key):
    from ._vertcoord import geos_etai_pressure, geos_etam_pressure
    from ._vertcoord import geos_hyam, geos_hybm
    if key == 'hyai':
        data = self.Ap
        dims = ('layer_bounds', )
        dtype = data.dtype.char
        if dims[0] not in self.dimensions:
            self.createDimension(dims[0], data.size)
        kwds = dict(units="hPa",
                    long_name="hybrid A coefficient at layer interfaces",
                    note="unit consistent with GEOS-Chem pressure outputs")
    elif key == 'hyam':
        data = geos_hyam[self.vertgrid]
        dims = ('layer',)
        dtype = data.dtype.char
        if dims[0] not in self.dimensions:
            self.createDimension(dims[0], data.size)
        kwds = dict(units="hPa",
                    long_name="hybrid A coefficient at layer midpoints",
                    note="unit consistent with GEOS-Chem pressure outputs")
    elif key == 'hybi':
        data = self.Bp
        dims = ('layer_bounds',)
        dtype = data.dtype.char
        if dims[0] not in self.dimensions:
            self.createDimension(dims[0], data.size)
        kwds = dict(
            units="1", long_name="hybrid B coefficient at layer interfaces")
    elif key == 'etai_pressure':
        data = geos_etai_pressure[self.vertgrid]
        dtype = data.dtype.char
        dims = ('layer_bounds',)
        kwds = dict(units='hPa', long_name='eta levels at interfaces')
        if dims[0] not in self.dimensions:
            self.createDimension(dims[0], data.size)
    elif key == 'etam_pressure':
        data = geos_etam_pressure[self.vertgrid]
        dtype = data.dtype.char
        dims = ('layer',)
        kwds = dict(units='hPa', long_name='eta levels at mid points')
        if dims[0] not in self.dimensions:
            self.createDimension(dims[0], data.size)
    elif key == 'hybm':
        data = geos_hybm[self.vertgrid]
        dims = ('layer', )
        dtype = data.dtype.char
        if dims[0] not in self.dimensions:
            self.createDimension(dims[0], data.size)
        kwds = dict(
            units="1", long_name="hybrid B coefficient at layer midpoints")
    tmpvar = self.createVariable(key, dtype, dims)
    nlay = tmpvar.size

    for p, v in kwds.items():
        setattr(tmpvar, p, v)

    tmpvar[:] = data[0:nlay]


def add_lat(self, key, example):
    yres = self.modelres[1]
    if self.halfpolar == 1:
        data = np.concatenate(
            [[-90.], np.arange(-90. + yres / 2., 90., yres), [90.]])
    else:
        data = np.arange(-90, 90 + yres, yres)
    dims = ('latitude',)
    dtype = 'f'
    kwds = dict(standard_name="latitude", long_name="latitude",
                units="degrees_north", base_units="degrees_north", axis="Y")
    if key == 'latitude':
        data = data[:-1] + np.diff(data) / 2.
        kwds['bounds'] = 'latitude_bounds'
    else:
        dims += ('nv',)
        data = data.repeat(2, 0)[1:-1].reshape(-1, 2)
    sj = getattr(example, 'STARTJ', 0)
    data = data[sj:sj + example.shape[2]]
    var = self.createVariable(key, dtype, dims)
    var[:] = data
    for p, v in kwds.items():
        setattr(var, p, v)


def add_lon(self, key, example):
    xres = self.modelres[0]
    i = np.arange(0, 360 + xres, xres)
    data = i - (180 + xres / 2. * self.center180)
    dims = ('longitude',)
    dtype = 'f'
    kwds = dict(standard_name="longitude", long_name="longitude",
                units="degrees_east", base_units="degrees_east", axis="X")
    if key == 'longitude':
        data = data[:-1] + np.diff(data) / 2.
        kwds['bounds'] = 'longitude_bounds'
    else:
        dims += ('nv',)
        data = data.repeat(2, 0)[1:-1].reshape(-1, 2)
    si = getattr(example, 'STARTI', 0)
    data = data[si:si + example.shape[3]]
    var = self.createVariable(key, dtype, dims)
    var[:] = data
    for p, v in kwds.items():
        setattr(var, p, v)


class gcvar(object):
    def __init__(self, key, parent):
        self._key = key
        self._name = key
        self._parent = parent
        self._data = None
        start, end, dim = list(self._parent._outpos[self._key].values())[0]
        ntimes = len(self._parent._outpos[self._key])
        nlevels = dim[2]
        nlatitudes = dim[1]
        nlongitudes = dim[0]
        timedim = 'time'
        laydim = 'layer%d' % nlevels
        self.dimensions = (timedim, laydim, 'latitude', 'longitude')
        self.shape = (ntimes, nlevels, nlatitudes, nlongitudes)
        dht = _datablock_header_type
        self._header = self._parent._data[start:start +
                                          _hdr_size].view(dht)
        self.category = self._header['category'][0].strip()
        self.tracerid = self._header['tracerid'][0]
        self.base_units = self._header['base_units'][0]
        self.catoffset = [row['offset']
                          for row in self._parent._ddata
                          if row['category'] == self.category][0]
        self.noscale = self._parent.noscale
        STARTK, STARTJ, STARTI = self._header['start'][0][::-1] - 1
        self.STARTK, self.STARTJ, self.STARTI = STARTK, STARTJ, STARTI
        if hasattr(self.category, 'decode'):
            self.category = self.category.decode()
        self.cattracerid = self.catoffset + self.tracerid
        props = ([row for row in self._parent._tdata
                  if row['tracerid'] == self.cattracerid] +
                 [row for row in self._parent._tdata
                  if row['tracerid'] == self.tracerid])[0]
        for pk in props.dtype.names:
            if pk == 'tracerid':
                continue
            pv = props[pk]
            if hasattr(pv, 'decode'):
                pv = pv.decode()
            setattr(self, pk, pv)

    def __getattr__(self, k):
        try:
            return object.__getattribute__(self, k)
        except Exception:
            return object.__getattribute__(self[:], k)

    def __getitem__(self, k):
        if self._data is None:
            pieces = []
            startenddims = self._parent._outpos[self._key]
            for (tstart, tend), (start, end, dim) in startenddims.items():
                # nelem = end - start - _hdr_size
                datat = np.dtype([('header', _datablock_header_type),
                                  ('bpad', '>i'),
                                  ('data', '>f', dim[::-1]),
                                  ('epad', '>i')])

                pieces.append(self._parent._data[start:end])
            tmpdata = np.concatenate(pieces, axis=0).view(datat)
            if self.noscale:
                self._data = tmpdata['data']
            else:
                self._data = tmpdata['data'] * self.scale
            self._tau0 = tmpdata['header']['tau0']
            self._tau1 = tmpdata['header']['tau1']
        return self._data.__getitem__(k)

    def ncattrs(self):
        return tuple(self.__dict__.keys())


class bpch2(bpch_base):
    @classmethod
    def isMine(cls, path, *args, **kwds):
        isbpch = oldbpch.isMine(path)
        if not isbpch:
            return False

        try:
            oldbpch(path)
            return False
        except Exception:
            return True

    @classmethod
    def from_ncf(cls, infile):
        outf = bpch_base()
        for pk in infile.ncattrs():
            pv = getattr(infile, pk)
            setattr(outf, pk, pv)

        for dk, dv in infile.dimensions.items():
            outf.copyDimension(dv, key=dk)

        for vk, vv in infile.variables.items():
            outf.copyVariable(vv, key=vk)

        return outf

    def __init__(self, path, nogroup=False, noscale=False,
                 vertgrid='GEOS-5-REDUCED'):
        self.noscale = noscale
        self._nogroup = nogroup
        from ._vertcoord import geos_hyai, geos_hybi
        self.vertgrid = vertgrid
        # Ap [hPa]
        self.Ap = geos_hyai[self.vertgrid]

        # Bp [unitless]
        self.Bp = geos_hybi[self.vertgrid]

        self._gettracerinfo(path)
        self._getdiaginfo(path)
        self._data = data = np.memmap(path, mode='r', dtype='uint8')
        header = data[:136].copy().view(_general_header_type)
        self.ftype = header[0][1]
        self.toptitle = header[0][4]

        offset = 136

        outpos = self._outpos = OrderedDict()

        while offset < data.size:
            tmp_hdr = data[offset:offset +
                           _hdr_size].view(_datablock_header_type)
            cat = tmp_hdr['category'][0].strip()
            if hasattr(cat, 'decode'):
                cat = cat.decode()
            trac = str(tmp_hdr['tracerid'][0])
            if hasattr(trac, 'decode'):
                trac = trac.decode()

            key = cat + '_' + trac
            tau0, tau1 = tmp_hdr['tau0'][0], tmp_hdr['tau1'][0]
            dim = tuple(tmp_hdr['dim'][0].tolist())
            skip = tmp_hdr['skip'][0]
            outpos.setdefault(key, OrderedDict())[(
                tau0, tau1)] = offset, offset + _hdr_size + skip, dim
            offset += skip + _hdr_size

        modelname, modelres, halfpolar, center180 = tmp_hdr[0].tolist()[1:5]
        self.modelname = modelname
        self.modelres = modelres
        self.halfpolar = halfpolar
        self.center180 = center180

        tmpvariables = self._gcvars = OrderedDict()
        for key in outpos:
            tmpvar = gcvar(key, self)
            if nogroup is True:
                tmpkey = str(tmpvar.shortname)
            elif nogroup is False:
                tmpkey = tmpvar.category + str('_') + tmpvar.shortname
            elif tmpvar.category in nogroup:
                tmpkey = str(tmpvar.shortname)
            else:
                tmpkey = tmpvar.category + str('_') + tmpvar.shortname

            tmpvariables[tmpkey] = tmpvar
        self.modelres = tmpvar._header['modelres'][0]
        self.halfpolar = tmpvar._header['halfpolar'][0]
        self.center180 = tmpvar._header['center180'][0]
        ntimes = [tmpvar.shape[0] for tmpkey, tmpvar in tmpvariables.items()]
        levels = [tmpvar.shape[1] for tmpkey, tmpvar in tmpvariables.items()]
        latitudes = [tmpvar.shape[2]
                     for tmpkey, tmpvar in tmpvariables.items()]
        longitudes = [tmpvar.shape[3]
                      for tmpkey, tmpvar in tmpvariables.items()]
        self._maxntimes = maxntimes = max(ntimes)
        self.createDimension('time', maxntimes)
        for time in set(ntimes):
            if time != maxntimes:
                self.createDimension('time%d' % time, time)
        self.createDimension('layer', min(max(levels), self.Ap.size - 1))
        for layer in set(levels):
            self.createDimension('layer%d' % layer, layer)
        self.createDimension('latitude', max(latitudes))
        self.createDimension('longitude', max(longitudes))
        self.createDimension('nv', 2)
        self.variables = PseudoNetCDFVariables(
            self._addvar, list(tmpvariables))
        key = list(tmpvariables)[0]
        tmpvar = self.variables[key]
        add_lat(self, 'latitude', tmpvar)
        add_lat(self, 'latitude_bounds', tmpvar)
        add_lon(self, 'longitude', tmpvar)
        add_lon(self, 'longitude_bounds', tmpvar)
        add_vert(self, 'hyam')
        add_vert(self, 'hyai')
        add_vert(self, 'hybm')
        add_vert(self, 'hybi')
        add_vert(self, 'etai_pressure')
        add_vert(self, 'etam_pressure')
        for k, v in [('tau0', tmpvar._tau0),
                     ('time', tmpvar._tau0),
                     ('tau1', tmpvar._tau1)]:
            tvar = self.createVariable(k, 'i', ('time',))
            tvar.units = 'hours since 1985-01-01 00:00:00 UTC'
            tvar[:] = v[:]

    def _addvar(self, key):
        tmpvar = self._gcvars[key]
        if tmpvar.shape[0] == self._maxntimes:
            timedim = 'time'
        else:
            timedim = 'time%d' % tmpvar.shape[0]
        laydim = 'layer%d' % tmpvar.shape[1]
        outdims = (timedim, laydim, 'latitude', 'longitude')
        var = self.createVariable(
            key, tmpvar.dtype.char, outdims, values=tmpvar)
        for k in tmpvar.ncattrs():
            if k in ('_name', 'shape', 'dimensions'):
                continue
            setattr(var, k, getattr(tmpvar, k))
        return var

    def _gettracerinfo(self, path):
        tpath = os.path.join(os.path.dirname(path), 'tracerinfo.dat')
        if not os.path.exists(tpath):
            tpath = 'tracerinfo.dat'
        self._tdata = np.recfromtxt(tpath, dtype=None, comments='#', names=[
                                    'shortname', 'fullname', 'kgpermole',
                                    'carbon', 'tracerid', 'scale', 'units'],
                                    delimiter=[9, 30, 10, 3, 9, 10, 41],
                                    autostrip=True)

    def _getdiaginfo(self, path):
        dpath = os.path.join(os.path.dirname(path), 'diaginfo.dat')
        if not os.path.exists(dpath):
            dpath = 'diaginfo.dat'
        self._ddata = np.recfromtxt(dpath, dtype=None, comments='#', names=[
                                    'offset', 'category', 'comment'],
                                    delimiter=[9, 40, 100], autostrip=True)

# OFFSET    (I8 )  Constant to add to tracer numbers in order to distinguish
#                  for the given diagnostic category, as stored in file
#                  "tracerinfo.dat".  OFFSET may be up to 8 digits long.
#  --       (1X )  1-character spacer
# CATEGORY  (A40)  Category name for CTM diagnostics. NOTE: The category name
#                  can be up to 40 chars long, but historically the GEOS-CHEM
#                  and GISS models have used an 8-character category name.
# COMMENT   (A  )  Descriptive comment string
#
#  --       (1X )  1-character spacer


class TestMemmaps(unittest.TestCase):
    def setUp(self):
        from PseudoNetCDF.testcase import geoschemfiles_paths
        self.bpchpath = geoschemfiles_paths['bpch']

    def testNCF2BPCH(self):
        bpchfile = bpch2(self.bpchpath, noscale=True)
        from PseudoNetCDF.pncgen import pncgen
        pncgen(bpchfile, self.bpchpath + '.check', inmode='r',
               outmode='w', format='bpch', verbose=0)
        orig = open(self.bpchpath, 'rb').read()
        new = open(self.bpchpath + '.check', 'rb').read()
        assert(orig == new)
        os.remove(self.bpchpath + '.check')

    def testBPCH2(self):
        bpchfile = bpch2(self.bpchpath)
        ALD2 = bpchfile.variables['IJ-AVG-$_ALD2']
        ALD2_check = np.array([1.60520077e-02, 1.82803553e-02, 2.00258084e-02,
                               2.01461259e-02, 1.84865110e-02, 2.49667447e-02,
                               2.73083989e-02, 2.87465211e-02, 2.89694592e-02,
                               2.87686456e-02, 2.87277419e-02, 3.08121163e-02,
                               3.22086290e-02, 3.35262120e-02, 3.41329686e-02,
                               3.05218045e-02, 3.30278911e-02, 3.58164124e-02,
                               3.93186994e-02, 4.15412188e-02, 1.60520077e-02,
                               1.82803553e-02, 2.00258084e-02, 2.01461259e-02,
                               1.84865110e-02, 2.49667447e-02, 2.73083989e-02,
                               2.87465211e-02, 2.89694592e-02, 2.87686456e-02,
                               2.87277419e-02, 3.08121163e-02, 3.22086290e-02,
                               3.35262120e-02, 3.41329686e-02, 3.05218045e-02,
                               3.30278911e-02, 3.58164124e-02, 3.93186994e-02,
                               4.15412188e-02, 1.60520077e-02, 1.82803553e-02,
                               2.00258084e-02, 2.01461259e-02, 1.84865110e-02,
                               2.49667447e-02, 2.73083989e-02, 2.87465211e-02,
                               2.89694592e-02, 2.87686456e-02, 2.87277419e-02,
                               3.08121163e-02, 3.22086290e-02, 3.35262120e-02,
                               3.41329686e-02, 3.05218045e-02, 3.30278911e-02,
                               3.58164124e-02, 3.93186994e-02,
                               4.15412188e-02]).reshape(ALD2.shape)
        np.testing.assert_allclose(ALD2, ALD2_check)
        hyai = np.array([0.0, 0.04804826, 6.593752,
                         13.1348, 19.61311, 26.09201, 32.57081, 38.98201,
                         45.33901, 51.69611, 58.05321, 64.36264, 70.62198,
                         78.83422, 89.09992, 99.36521, 109.1817, 118.9586,
                         128.6959, 142.91, 156.26, 169.609, 181.619, 193.097,
                         203.259, 212.15, 218.776, 223.898, 224.363, 216.865,
                         201.192, 176.93, 150.393, 127.837, 108.663, 92.36572,
                         78.51231, 56.38791, 40.17541, 28.36781, 19.7916,
                         9.292942, 4.076571, 1.65079, 0.6167791, 0.211349,
                         0.06600001, 0.01], 'f')
        np.testing.assert_allclose(bpchfile.variables['hyai'], hyai)
        hybi = np.array([1.0, 0.984952, 0.963406, 0.941865,
                         0.920387, 0.898908, 0.877429, 0.856018, 0.8346609,
                         0.8133039, 0.7919469, 0.7706375, 0.7493782, 0.721166,
                         0.6858999, 0.6506349, 0.6158184, 0.5810415, 0.5463042,
                         0.4945902, 0.4437402, 0.3928911, 0.3433811,
                         0.2944031, 0.2467411, 0.2003501, 0.1562241,
                         0.1136021, 0.06372006, 0.02801004, 0.006960025,
                         8.175413e-09, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], 'f')
        np.testing.assert_allclose(bpchfile.variables['hybi'], hybi)
        etaip = np.array([1.01325000e+03, 9.98050662e+02,
                          9.82764882e+02, 9.67479511e+02, 9.52195238e+02,
                          9.36910541e+02, 9.21625744e+02, 9.06342248e+02,
                          8.91059167e+02, 8.75776287e+02, 8.60493406e+02,
                          8.45211087e+02, 8.29929441e+02, 8.09555669e+02,
                          7.84087994e+02, 7.58621022e+02, 7.33159694e+02,
                          7.07698900e+02, 6.82238631e+02, 6.44053520e+02,
                          6.05879758e+02, 5.67705907e+02, 5.29549900e+02,
                          4.91400941e+02, 4.53269420e+02, 4.15154739e+02,
                          3.77070069e+02, 3.39005328e+02, 2.88927351e+02,
                          2.45246173e+02, 2.08244245e+02, 1.76930008e+02,
                          1.50393000e+02, 1.27837000e+02, 1.08663000e+02,
                          9.23657200e+01, 7.85123100e+01, 5.63879100e+01,
                          4.01754100e+01, 2.83678100e+01, 1.97916000e+01,
                          9.29294200e+00, 4.07657100e+00, 1.65079000e+00,
                          6.16779100e-01, 2.11349000e-01, 6.60000100e-02,
                          1.00000000e-02])
        np.testing.assert_allclose(bpchfile.variables['etai_pressure'], etaip)

    def runTest(self):
        pass


if __name__ == '__main__':
    bpchpath = ('/Users/barronh/Development/pseudonetcdf/src/PseudoNetCDF/' +
                'testcase/geoschemfiles/test.bpch')
    f = bpch2(bpchpath)
    var = f.variables['IJ-AVG-$_Ox']
    print(var[:].mean())



#!/usr/bin/env python
# Run this script with pnc options
from __future__ import print_function

import os

from PseudoNetCDF.pncload import PNCConsole
from PseudoNetCDF.pncparse import pncparse, getparser
from PseudoNetCDF.coordutil import gettimes
import warnings
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

warn = warnings.warn
Normalize = matplotlib.colors.Normalize

LogNorm = matplotlib.colors.LogNorm
SymLogNorm = matplotlib.colors.SymLogNorm
BoundaryNorm = matplotlib.colors.BoundaryNorm

LogFormatter = matplotlib.ticker.LogFormatter
ScalarFormatter = matplotlib.ticker.ScalarFormatter

fig = plt.figure()
ax = fig.add_subplot(111)


def make1d(ifile, args):
    if len(args.figure_keywords) > 0:
        plt.setp(fig, **args.figure_keywords)
    if len(args.axes_keywords) > 0:
        plt.setp(ax, **args.axes_keywords)
    for fi, ifile in enumerate(ifiles):
        variables = args.variables
        if variables is None:
            variables = [key for key, var in ifile.variables.items()
                         if var.ndim == 2]
        if len(variables) == 0:
            raise ValueError(
                'Unable to heuristically determin plottable variables; use -v to specify variables for plotting')
        for varkey in variables:
            if not args.overlay:
                ax.cla()
            var = ifile.variables[varkey]
            vals = var[:]
            if args.squeeze:
                vals = vals.squeeze()
            label = getattr(var, 'standard_name', varkey).strip()
            varunit = getattr(var, 'units', 'unknown').strip()
            print(varkey, sep='')
            dimkey = var.dimensions[0]
            ax.set_xlabel(dimkey)
            if args.time:
                x = gettimes(ifile)
            elif dimkey in ifile.variables:
                x = ifile.variables[var.dimensions[0]][:]
            else:
                x = np.arange(var.shape[0])

            patches = ax.plot(x, vals, label=label)
            ax.set_ylabel(varunit)
            plt.legend()
            fmt = 'png'
            outpath = args.outpath
            if len(ifiles) > 1:

                lstr = str(fi).rjust(len(str(len(ifiles))), '0')
            else:
                lstr = ''

            figpath = os.path.join(outpath + varkey + lstr + '.' + fmt)
            if args.interactive:
                csl = PNCConsole(locals=globals())
                csl.interact()

            fig.savefig(figpath)
            if args.verbose > 0:
                print('Saved fig', figpath)


if __name__ == '__main__':
    parser = getparser(has_ofile=True, plot_options=True, interactive=True)
    parser.add_argument('--no-squeeze', dest='squeeze', default=True, action='store_false',
                        help='Squeeze automatically removes singleton dimensions; disabling requires user to remove singleton dimensions with --remove-singleton option')
    parser.add_argument('--time', action='store_true',
                        help='Use time object as x-axis')
    ifiles, args = pncparse(
        has_ofile=True, plot_options=True, interactive=True, parser=parser)
    make1d(ifiles, args)



__all__ = ['arlpackedbit', 'l100', 'arlpardump', 'arlconcdump', 'arltrajdump']

from ._arl import arlpackedbit
from ._l100 import l100
from ._pdump import arlpardump
from ._cdump import arlconcdump
from ._tdump import arltrajdump



__all__ = ['height_pressure']
__doc__ = """
.. _Memmap
:mod:`Memmap` -- height/pressure Memmap interface
=================================================

.. module:: Memmap
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map for CAMx
              height/pressure files.  See PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

# Distribution packages
import unittest
import struct

# Site-Packages
from numpy import zeros, array, memmap

# This Package modules
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariable
from PseudoNetCDF.sci_var import PseudoNetCDFVariables
from PseudoNetCDF.ArrayTransforms import ConvertCAMxTime

# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class height_pressure(PseudoNetCDFFile):
    """
    height_pressure provides a PseudoNetCDF interface for CAMx
    height_pressure files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> height_pressure_path = 'camx_height_pressure.bin'
        >>> rows,cols = 65,83
        >>> hpf = height_pressure(height_pressure_path,rows,cols)
        >>> hpf.variables.keys()
        ['TFLAG', 'HGHT', 'PRES']
        >>> v = hpf.variables['V']
        >>> tflag = hpf.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> hpf.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    id_fmt = 'fi'
    data_fmt = 'f'

    def __init__(self, rf, rows=None, cols=None):
        self.__memmap = memmap(rf, '>f', 'r', offset=0)
        rowsXcols = self.__memmap[0].view('i') // 4 - 2
        record_length = rowsXcols + 4
        records = self.__memmap.size // record_length

        times = self.__memmap.reshape(records, record_length)[:, 1:3]
        self.STIME, self.SDATE = times[0]
        for i, (t, d) in enumerate(times):
            if (t, d) != (self.STIME, self.SDATE):
                break
        self.SDATE = self.SDATE.view('i')
        self.createDimension('LAY', i / 2)
        self.createDimension('TSTEP', times.shape[0] / i)

        if rows is None and cols is None:
            rows = rowsXcols
            cols = 1
        elif rows is None:
            rows = rowsXcols / cols
        elif cols is None:
            cols = rowsXcols / rows
        else:
            if cols * rows != rowsXcols:
                raise ValueError(("The product of cols (%d) and rows (%d) " +
                                  "must equal cells (%d)") % (
                                 cols, rows, rowsXcols))

        self.createDimension('ROW', rows)
        self.createDimension('COL', cols)
        self.createDimension('DATE-TIME', 2)
        self.createDimension('VAR', 2)

        self.NROWS = rows
        self.NCOLS = cols
        self.NLAYS = len(self.dimensions['LAY'])
        self.NVARS = 2
        self.NTHIK = 1

        setattr(self, 'VAR-LIST', 'HGHT'.ljust(16) + 'PRES'.ljust(16))

        self.variables = PseudoNetCDFVariables(
            self.__var_get, ['HGHT', 'PRES', 'TFLAG'])

    def __var_get(self, key):
        lays = len(self.dimensions['LAY'])
        times = len(self.dimensions['TSTEP'])
        rows = len(self.dimensions['ROW'])
        cols = len(self.dimensions['COL'])
        hght = 1
        pres = 2
        time = 3
        date = 4
        out_idx = zeros(self.__memmap.shape, dtype='b').reshape(
            times, lays, 2, rows * cols + 4)
        out_idx[:, :, 0, 3:-1] = hght
        out_idx[:, :, 1, 3:-1] = pres
        out_idx[:, :, :, 1] = time
        out_idx[:, :, :, 2] = date
        out_idx = out_idx.ravel()
        buf = self.__memmap[out_idx == 0].reshape(lays * 2 * times, 2)
        if not (buf[:, 0] == buf[:, 1]).all():
            raise ValueError("Buffer")
        tmpvals = self.__memmap[out_idx == 1].reshape(times, lays, rows, cols)
        v = self.variables['HGHT'] = PseudoNetCDFVariable(
            self, 'HGHT', 'f', ('TSTEP', 'LAY', 'ROW', 'COL'), values=tmpvals)
        v.units = 'm'
        v.long_name = 'HGHT'.ljust(16)
        v.var_desc = 'Top Height'
        tmpvals = self.__memmap[out_idx == 2].reshape(times, lays, rows, cols)
        v = self.variables['PRES'] = PseudoNetCDFVariable(
            self, 'PRES', 'f', ('TSTEP', 'LAY', 'ROW', 'COL'), values=tmpvals)
        v.units = 'hPA'
        v.long_name = 'PRES'.ljust(16)
        v.var_desc = 'Pressure at center'
        tslice = slice(None, None, len(self.dimensions['LAY']) * 2)
        dateblock = self.__memmap[out_idx == 4][tslice].view('>i')
        hourblock = self.__memmap[out_idx == 3][tslice]
        self.variables['TFLAG'] = ConvertCAMxTime(dateblock, hourblock,
                                                  len(self.dimensions['VAR']))

        return self.variables[key]


class TestMemmap(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testHP(self):
        import PseudoNetCDF.testcase
        hpfile = height_pressure(
            PseudoNetCDF.testcase.camxfiles_paths['height_pressure'], 4, 5)
        hpfile.variables['TFLAG']
        checkv = array([3.38721924e+01, 3.40657959e+01, 3.41392822e+01,
                        3.42358398e+01, 3.42543945e+01, 3.38868408e+01,
                        3.40622559e+01, 3.42358398e+01, 3.44768066e+01,
                        3.46112061e+01, 3.37558594e+01, 3.39323730e+01,
                        3.42663574e+01, 3.46854248e+01, 3.48144531e+01,
                        3.39472656e+01, 3.41900635e+01, 3.46160889e+01,
                        3.48209229e+01, 3.47874756e+01, 6.78652344e+01,
                        6.82532959e+01, 6.84020996e+01, 6.85950928e+01,
                        6.86304932e+01, 6.78945312e+01, 6.82465820e+01,
                        6.85941162e+01, 6.90783691e+01, 6.93474121e+01,
                        6.76313477e+01, 6.79859619e+01, 6.86558838e+01,
                        6.94960938e+01, 6.97552490e+01, 6.80159912e+01,
                        6.85028076e+01, 6.93570557e+01, 6.97674561e+01,
                        6.97009277e+01, 1.01980713e+02, 1.02563843e+02,
                        1.02787109e+02, 1.03077759e+02, 1.03131104e+02,
                        1.02022949e+02, 1.02553101e+02, 1.03076904e+02,
                        1.03804565e+02, 1.04208984e+02, 1.01628662e+02,
                        1.02162842e+02, 1.03169922e+02, 1.04433838e+02,
                        1.04823120e+02, 1.02206909e+02, 1.02940186e+02,
                        1.04224609e+02, 1.04841797e+02, 1.04740723e+02,
                        3.38721924e+01, 3.40657959e+01, 3.41392822e+01,
                        3.42358398e+01, 3.42543945e+01, 3.38868408e+01,
                        3.40622559e+01, 3.42358398e+01, 3.44768066e+01,
                        3.46112061e+01, 3.37558594e+01, 3.39323730e+01,
                        3.42663574e+01, 3.46854248e+01, 3.48144531e+01,
                        3.39472656e+01, 3.41900635e+01, 3.46160889e+01,
                        3.48209229e+01, 3.47874756e+01, 6.78652344e+01,
                        6.82532959e+01, 6.84020996e+01, 6.85950928e+01,
                        6.86304932e+01, 6.78945312e+01, 6.82465820e+01,
                        6.85941162e+01, 6.90783691e+01, 6.93474121e+01,
                        6.76313477e+01, 6.79859619e+01, 6.86558838e+01,
                        6.94960938e+01, 6.97552490e+01, 6.80159912e+01,
                        6.85028076e+01, 6.93570557e+01, 6.97674561e+01,
                        6.97009277e+01, 1.01980713e+02, 1.02563843e+02,
                        1.02787109e+02, 1.03077759e+02, 1.03131104e+02,
                        1.02022949e+02, 1.02553101e+02, 1.03076904e+02,
                        1.03804565e+02, 1.04208984e+02, 1.01628662e+02,
                        1.02162842e+02, 1.03169922e+02, 1.04433838e+02,
                        1.04823120e+02, 1.02206909e+02, 1.02940186e+02,
                        1.04224609e+02, 1.04841797e+02, 1.04740723e+02],
                       dtype='f').reshape(2, 3, 4, 5)
        self.assertTrue((hpfile.variables['HGHT'] == checkv).all())

    def testNCF2HP(self):
        import PseudoNetCDF.testcase
        from PseudoNetCDF.pncgen import pncgen
        import os
        inpath = PseudoNetCDF.testcase.camxfiles_paths['height_pressure']
        outpath = inpath + '.check'
        infile = height_pressure(inpath, 4, 5)
        pncgen(infile, outpath, format='camxfiles.height_pressure')
        orig = open(inpath, 'rb').read()
        new = open(outpath, 'rb').read()
        assert(orig == new)
        os.remove(outpath)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# Run this script with pnc options
import numpy as np


def pncwindrose(ifiles, args):
    if len(ifiles) > 2:
        raise ValueError(
            'Requires one or two input files (received %d). To combine, use --stack' % len(ifiles))

    ws = np.ma.masked_invalid(
        eval(args.windspeed, None, ifiles[0].variables)[:])
    wd = np.ma.masked_invalid(eval(args.winddir, None, ifiles[0].variables)[:])
    if len(ifiles) == 2:
        maskvar = eval(args.maskvar, None, ifiles[1].variables)
        ws = np.ma.masked_where(maskvar, ws[:])
        wd = np.ma.masked_where(maskvar, wd[:])

    windrose(wd[:], ws[:], args)


def windrose(wd, ws, args):
    from matplotlib import pyplot as plt
    plt.rcParams['text.usetex'] = False
    width = args.binwidth
    halfwidth = width / 2.
    fig = plt.figure(figsize=(6, 7))
    ax = fig.add_axes([0.1, 0.2, 0.8, 0.7], polar=True)
    ax.set_theta_offset(np.radians(90))
    ax.set_theta_direction(-1)
    ubs = args.bounds[1:]
    lbs = args.bounds[:-1]
    ubcs = plt.cm.jet(np.arange(len(ubs), dtype='f') / ubs.size)
    color_dict = dict(zip(ubs, ubcs))
    mask = np.ma.getmaskarray(ws) | np.ma.getmaskarray(
        wd) | (ws < args.bounds[0]) | (ws > args.bounds[-1])
    ws = np.ma.masked_where(mask, ws).compressed()
    wd = np.ma.masked_where(mask, wd).compressed()
    fig.suptitle('%s Wind Rose (%d)' % (args.title, ws.size), size=20)
    if args.fromnorth:
        # NNE, WNE, WSE, SSE, SSW, WSW, WNW, NNW
        thetas = ((np.radians((wd) // width * width)).astype('f') %
                  (2 * np.pi)).astype('f')
    else:
        # N, NE, E, SE, S, SW, W, NW
        thetas = (np.radians((wd.astype('d') + halfwidth) //
                             width * width - halfwidth) % (2 * np.pi)).astype('f')
    # bounds = np.degrees(thetas)[:, None] + np.array([0, width])[None, :] % 360
    uthetas = np.unique(thetas)
    nwinds = float(ws.size)
    for theta in uthetas:
        tws = ws[thetas == theta]
        for ubi, ub in enumerate(ubs[::-1]):
            ubc = color_dict[ub]
            pct = (tws < ub).sum() / nwinds * 100.
            ax.bar(theta, pct, width=np.radians(width), bottom=0.0, color=ubc)
    labels = [(('<%s m/s' % ub) if ub != np.inf else ('>%s m/s' % lb))
              for lb, ub in zip(lbs, ubs)]
    rectangles = [plt.Rectangle((0, 0), 1, 1, color=ubc) for ubc in ubcs]
    fig.legend(rectangles, labels, ncol=3,
               bbox_to_anchor=(.5, 0.025), loc='lower center')
    ticks = np.linspace(0, args.maxpct, 6)[1:]
    labels = ['%s%%' % i for i in ticks]
    plt.yticks(ticks, labels)
    ax.set_rmax(args.maxpct)
    ax.set_clip_on(False)
    figpath = args.outpath
    fig.savefig(figpath)
    if args.verbose > 0:
        print('Saved fig', figpath)
    plt.close(fig)


if __name__ == '__main__':
    from PseudoNetCDF.pncparse import getparser, pncparse
    parser = getparser(has_ofile=True, plot_options=False, interactive=False)
    parser.add_argument('--wind-speed', dest='windspeed',
                        default='WS', help='Wind speed variable')
    parser.add_argument('--mask-var', dest='maskvar',
                        default='WS', help='Variable from file two for masking')
    parser.add_argument('--wind-dir', dest='winddir',
                        default='WD', help='Wind direction variable')
    parser.add_argument('--title', default=None, help='Title for wind rose')
    parser.add_argument('--binwidth', dest='binwidth', type=float,
                        default=45., help='Direction bin widths in degrees.')
    parser.add_argument('--max-pct', dest='maxpct', type=float,
                        default=50., help='Maximum percent on plot')
    parser.add_argument('--fromnorth', dest='fromnorth', action='store_true',
                        default=False, help='Bins should start from north (i.e., not bound north)')
    parser.add_argument('--bounds', dest='bounds', type=lambda x: np.array(eval(x)),
                        default=np.array([1, 2, 3, 5, 10, np.inf]), help='Boundaries for wind speed')
    parser.epilog = """
Example:
    $ pncwindrose.py -s time,600,None --max-pct=35 --wind-dir wd --wind-speed ws  --mask-var="ws[:].mask.all(0)[None,:].repeat(ws.shape[0], 0)" --bounds="[0.5,1,2,3.5,4.5,6.5,np.inf]"  --title obs  Bogota_2012_s12345.nc df.nc oct_obs

Description:
    Create windrose for timestep 600 and on where wind-direction is name wd, and windspeed is named ws. Masking is applied from df.nc file where all times for ws are masked (useful for ignoring monitors). Values less than 0.5 will be masked.
"""
    ifiles, args = pncparse(has_ofile=True, plot_options=False,
                            interactive=False, args=None, parser=parser)
    out = pncwindrose(ifiles, args)



from __future__ import print_function
import re
import unittest

# site-packages
from numpy import array, zeros, newaxis
from PseudoNetCDF.sci_var import PseudoNetCDFFile as pncf

net_num_re = '(?<=\s)[-]?\d+[.]?\d+'
net_output_re = re.compile('^\w+(\s+' + net_num_re + '){24}$')
net_name_re = re.compile(
    '[+]|source|termination|production|Source|Termination|Production|' +
    'Oxidation|oxidation|Radical|radical')
net_time_dim_re = re.compile('^Ending Hour')
net_spc_name_re = re.compile('^\w+')
net_time_re = re.compile(net_num_re + '|daily|Daily')
net_num_re = re.compile(net_num_re)

sum_num_pattern = '([-]?\d+[.]\d+)'
sum_num_sep = '\s+'


def print_net_rxns(net_rxns, time='Daily'):
    nrxns = net_rxns.NET_RXN
    for nrxn in nrxns:
        print("%s: %s" % (nrxn, net_reaction(net_rxns, nrxn, time)))


class sum_reader(pncf):
    @classmethod
    def isMine(cls, *args, **kwds):
        return False

    def __init__(self, sumfile):
        if type(sumfile) == str:
            self.sumfile = open(sumfile, 'r')
        else:
            self.sumfile = sumfile
        if not 'Sum'.lower() in self.sumfile.readline().lower():
            raise IOError('Not a sum file')
        self.sumfile.seek(0, 0)
        time, var_names, values = self.parse()

        self.createDimension('TSTEP', len(time))

        for vn, vv in zip(var_names, values):
            tv = self.createVariable(vn, 'f', ('TSTEP',))
            tv[:] = vv

        tv = self.createVariable('TFLAG', 'S10', ('TSTEP',))
        tv[:] = time

    def parse(self):
        time = []
        var_names = []
        values = []
        lines = self.sumfile.readlines()
        for line in lines:
            line = line[:-1]
            if net_time_dim_re.match(line) is not None and time == []:
                time = list(set(line.split(' ')[2:]))[1:]
                time.sort()
            sum_num_re = re.compile(
                '(.*)\s+' + sum_num_sep.join([sum_num_pattern] * len(time)))

            if time != [] and sum_num_re.match(line) is None and line != '':
                category = line

            if time != [] and sum_num_re.match(line) is not None:
                grps = sum_num_re.match(line).groups()
                var_names.append(category.strip() + ': ' + grps[0].strip())
                values.append(map(float, grps[1:]))

        return time, var_names, values


def net_reaction(net_rxns, net_rxn, time='Daily'):
    spcs = net_rxns.SPECIES
    reactants = []
    products = []
    tmp = "%f %s"
    for spc in spcs:
        v = net_rxns(NET_RXN=net_rxn, TIME=time, SPECIES=spc)
        if v == 0:
            pass
        elif v > 0:
            products.extend((v, spc))
        elif v < 0:
            reactants.extend((-1 * v, spc))

    return ((" + ".join([tmp for i in reactants if type(i) == str]) + " <-> " +
             " + ".join([tmp for i in products if type(i) == str])) %
            tuple(reactants + products))


class ctb_reader(pncf):
    @classmethod
    def isMine(cls, *args, **kwds):
        return False

    def __init__(self, file):
        if type(file) == str:
            file = open(file)
        lines = file.readlines()
        parameters, daily, peak = self.parse(lines)

        self.createDimension('PARAMETERS', len(parameters))
        self.createDimension('PEAK_DAILY', 2)

        v = self.createVariable('CTB', 'f', ('PARAMETERS', 'PEAK_DAILY'))
        v[:] = array([daily, peak]).swapaxes(0, 1)
        v.units = 'ppbV'

    def parse(self, lines):
        lines = lines[16:]
        parameters, daily, hourly = [], [], []
        for l in lines:
            try:
                p, d, h = l[:-1].split('\t')
                parameters.append(p)
                daily.append(float(d))
                hourly.append(float(h))
            except Exception:
                break
        return parameters, daily, hourly


class net_reader(pncf):
    @classmethod
    def isMine(cls, *args, **kwds):
        return False

    def __init__(self, infile):
        self.spc = []
        self.nrxns = []
        self.time = []

        if type(infile) == str:
            infile = open(infile)
        lines = infile.readlines()
        self.parse(lines, False)
        self.createDimension('NET_RXN', len(self.nrxns))
        self.createDimension('SPECIES', len(self.spc))
        self.createDimension('TSTEP', len(self.time))
        self.createVariable('NET', 'f', ('TSTEP', 'NET_RXN', 'SPECIES'))
        self.parse(lines, True)

    def parse(self, lines, fill=False):
        for l in lines:
            if net_time_dim_re.match(l) is not None and self.time == []:
                self.time = net_time_re.findall(l)
            elif net_name_re.search(l) is not None:
                netrxn = l.strip()
                if netrxn not in self.nrxns:
                    self.nrxns.append(netrxn)
            elif net_output_re.match(l) is not None:
                spcn = net_spc_name_re.search(l).group()
                if spcn not in self.spc:
                    self.spc.append(spcn)
                if fill:
                    slc = self.variables['NET'][:, self.nrxns.index(
                        netrxn), self.spc.index(spcn)]
                    slc[:] = list(map(float, net_num_re.findall(l)))


class mrgaloft(pncf):
    import re
    # create regular expressions to read the input file..
    sci_not = '-?\d+.\d+[E,e][+,-]\d+'
    time_re = re.compile('Time =([0-9][0-9]0000)', re.IGNORECASE)
    irr_re = re.compile('\{\s*(\d+)\}\s+(' + sci_not + ')', re.IGNORECASE)
    ipr_re = re.compile('"\w+\s*"\s*(' + sci_not + '\s+)+', re.IGNORECASE)
    split_on_blanks_re = re.compile('[ ]+')

    @classmethod
    def isMine(cls, *args, **kwds):
        return False

    def __init__(self, mrgfile):
        if hasattr(mrgfile, 'seek') and hasattr(mrgfile, 'readline'):
            pass
        else:
            try:
                mrgfile = open(mrgfile, 'rb')
            except Exception as e:
                raise TypeError("mrgfile must be a path or a file; " + str(e))

        self.mrgfile = mrgfile
        self.__initfile()
        self.__get_spc_prc()
        self.__internalize()
        self.__initfile()

    def __get_spc_prc(self):
        # Read and discard time and irr
        self.read_time()
        self.read_irr()

        # Read ipr with species set to true
        # discard process information and internalize spc names
        self.spc, self.prc = self.read_ipr(True, True)[1:]
        self.Species = "".join([i.ljust(16) for i in self.spc])
        self.Process = "".join([i.ljust(16) for i in self.prc])

    def __initfile(self):
        # Move to start of file
        self.mrgfile.seek(0)
        # Internalize ipr and irr file names
        self.iprfilename = self.readline()[:-1]
        self.irrfilename = self.readline()[:-1]

    def __internalize(self):
        self.times = []
        self.irr = []
        self.ipr = []
        # Internalize all arrays
        for t, r, p in self:
            self.times.append(t)
            self.irr.append(r)
            self.ipr.append(p)
        self.createDimension('TSTEP', len(self.times))
        self.createDimension('RXNS', len(self.irr[0]))
        self.createDimension('VAR', 2)
        self.createDimension('DATE-TIME', 2)
        self.createDimension('SPECIES', len(self.ipr[0]))
        self.createDimension('PROCESS', len(self.ipr[0][0]))
        v = self.createVariable('IRR', 'f', ('TSTEP', 'RXNS'))
        v[:] = array(self.irr)
        v.units = 'ppb'
        v = self.createVariable('IPR', 'f', ('TSTEP', 'SPECIES', 'PROCESS'))
        v[:] = array(self.ipr)
        v.units = 'ppb'
        v = self.createVariable('TFLAG', 'I', ('TSTEP', 'VAR', 'DATE-TIME'))
        v[:, :, :] = array(
            (zeros(len(self.times), 'i'),
             self.times), 'i').swapaxes(0, 1)[:, newaxis, :]
        v[:, :, 1]

    def readline(self):
        return self.mrgfile.readline()

    def read_time(self):
        f = self.mrgfile
        line = f.readline()
        if line[0].strip() == '|':
            raise EOFError('EOF File: found |')

        # next line is first Time =
        ir_time = self.time_re.match(line).groups()[0]
        if ir_time is None:
            raise ValueError("ERROR:: in get_data read, did not find a time.")
        return int(ir_time)

    def read_irr(self):
        f = self.mrgfile
        # initialize ir and ip storage...
        ir_rates = []  # account for 1-origin of rates

        # read in the !"Rxn No"    "Int Rate" header
        line = f.readline()

        line = f.readline()

        # read the irr values
        while line[0] != ';':
            ir_value = self.irr_re.match(line).groups()[1]
            if ir_value is None:    # if this is a { n} n.nnnn line ...
                raise ValueError(
                    "Expecting irr formatted line (i.e. { n} n.nnnn line ...)")
            else:
                ir_rates.append(float(ir_value))
            line = f.readline()
        return ir_rates

    def read_ipr(self, spc=False, prc=False):
        f = self.mrgfile
        # read in the ! Species      Initial conc. ... header
        line = f.readline()[15:]
        r = range(len(line) // 17)
        ip_prc = [line[i * 17:i * 17 + 17].strip() for i in r]

        #
        # ip_rates = [ [ip_values], [ip_values], [ip_values], ... ]
        #   order of sub-lists 'ip_values' is order of iProcess id Number
        #   for this hour.
        line = f.readline()
        ip_rates = []
        ip_spc = []
        while line[0] != ';':
            ir_set = self.ipr_re.match(line)
            if ir_set is not None:
                ip_spc.append(line[1:9].strip())
                ip_set = self.split_on_blanks_re.split(line[11:-1].strip())
                ip_values = [float(v) for v in ip_set]
                ip_rates.append(ip_values)
            line = f.readline()
        if spc and prc:
            return ip_rates, ip_spc, ip_prc
        elif spc:
            return ip_rates, ip_spc
        elif prc:
            return ip_rates, ip_prc
        else:
            return ip_rates

    def __iter__(self):
        self.__initfile()
        while True:
            try:
                yield self.read_time(), self.read_irr(), self.read_ipr()
            except EOFError:
                raise StopIteration


class TestReaders(unittest.TestCase):
    def setUp(self):
        from PseudoNetCDF.testcase import net_balance_paths
        self.netfile = net_balance_paths['net_file']
        self.mrgfile = net_balance_paths['mrg_file']

    def testNet(self):
        net_reader(open(self.netfile))

    def testMrg(self):
        mrgfile = mrgaloft(open(self.mrgfile))
        irr = mrgfile.variables['IRR']
        ipr = mrgfile.variables['IPR']
        self.assertEqual(irr[0, 60], array(2.34061E-02, 'f')[...])
        self.assertEqual(irr[6, 60], array(5.21899E-04, 'f')[...])

        self.assertEqual(ipr[0, 6, 3], array(1.290572E+01, 'f')[...])
        self.assertEqual(ipr[6, 6, 3], array(1.497632E-01, 'f')[...])

    def runTest(self):
        pass


if __name__ == '__main__':
    unittest.main()



from __future__ import unicode_literals
__all__ = ['point_source']
__doc__ = """
.. _Memmap
:mod:`Memmap` -- point_source Memmap interface
============================================

.. module:: Memmap
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map for CAMx
              point_source files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@gmail.com>
.. modulecontributor:: Yoann Long (little endian support)
"""

# Distribution packages
import unittest
import struct

# Site-Packages
from numpy import zeros, array, memmap, dtype
import numpy as np
# This Package modules
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariable
from PseudoNetCDF.sci_var import PseudoNetCDFVariables
from PseudoNetCDF.ArrayTransforms import ConvertCAMxTime

# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class point_source(PseudoNetCDFFile):
    """
    point_source provides a PseudoNetCDF interface for CAMx
    point_source files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> point_source_path = 'camx_point_source.bin'
        >>> rows,cols = 65,83
        >>> point_sourcefile = point_source(point_source_path,rows,cols)
        >>> point_sourcefile.variables.keys()
        ['TFLAG', 'ETFLAG', 'TFLAG', 'XSTK', 'YSTK', 'HSTK', 'DSTK', 'TSTK',
         'VSTK', 'KCELL', 'FLOW', 'PLMHT', 'NSTKS', 'NO', 'NO2', ...]
        >>> tflag = point_sourcefile.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v = point_sourcefile.variables['XSTK']
        >>> v.dimensions
        ('NSTK',)
        >>> v.shape
        (38452,)
        >>> v = point_sourcefile.variables['NO2']
        >>> v.dimensions
        ('TSTEP', 'NSTK')
        >>> v.shape
        (25, 38452)
        >>> point_sourcefile.dimensions
        {'TSTEP': 25, 'NSTK': 38452}
    """
    @classmethod
    def isMine(cls, path):
        try:
            point_source(path)
            return True
        except Exception:
            return False

    def __init__(self, rf, endian='big', mode='r', **kwds):
        """
        Initialization included reading the header and learning
        about the format.

        see __readheader and __gettimestep() for more info
        """
        self.__rffile = rf

        ep = self.__endianprefix = dict(big='>', little='<')[endian]
        self.variables = {}
        for key, val in kwds.items():
            setattr(self, key, val)

        self.__memmap = memmap(self.__rffile, ep + 'f', mode)
        self.__emiss_hdr_fmt = dtype(dict(
            names=['SPAD', 'name', 'note', 'itzon', 'nspec', 'ibdate',
                   'btime', 'iedate', 'etime', 'EPAD'],
            formats=['i', '(10,4)S1', '(60,4)S1', 'i', 'i', 'i', 'f', 'i',
                     'f', 'i'])).newbyteorder(ep)
        self.__grid_hdr_fmt = dtype(dict(
            names=['SPAD', 'plon', 'plat', 'iutm', 'xorg', 'yorg', 'delx',
                   'dely', 'nx', 'ny', 'nz', 'iproj', 'istag', 'tlat1',
                   'tlat2', 'rdum5', 'EPAD'],
            formats=['i', 'f', 'f', 'i', 'f', 'f', 'f', 'f', 'i', 'i', 'i',
                     'i', 'i', 'f', 'f', 'f', 'i'])).newbyteorder(ep)
        self.__cell_hdr_fmt = dtype(dict(
            names=['SPAD', 'ione1', 'ione2', 'nx', 'ny', 'EPAD'],
            formats=['i', 'i', 'i', 'i', 'i', 'i'])).newbyteorder(ep)
        self.__spc_hdr_fmt = dtype('(10,4)' + 'S1').newbyteorder(ep)
        self.__time_hdr_fmt = dtype(dict(
            names=['SPAD', 'ibdate', 'btime', 'iedate', 'etime', 'EPAD'],
            formats=['i', 'i', 'f', 'i', 'f', 'i'])).newbyteorder(ep)
        self.__nstk_hdr_fmt = dtype(dict(
            names=['SPAD', 'ione', 'nstk', 'EPAD'],
            formats=['i', 'i', 'i', 'i'])).newbyteorder(ep)
        self.__stk_prop_fmt = dtype(dict(
            names=['XSTK', 'YSTK', 'HSTK', 'DSTK', 'TSTK', 'VSTK'],
            formats=['f', 'f', 'f', 'f', 'f', 'f'])).newbyteorder(ep)
        self.__stk_time_prop_fmt = dtype(dict(
            names=['IONE', 'ITWO', 'KCELL', 'FLOW', 'PLMHT'],
            formats=['i', 'i', 'f', 'f', 'f'])).newbyteorder(ep)

        self.__globalheader()
        varkeys = ['ETFLAG', 'TFLAG', 'XSTK', 'YSTK', 'HSTK', 'DSTK', 'TSTK',
                   'VSTK', 'IONE', 'ITWO', 'KCELL', 'FLOW', 'PLMHT',
                   'NSTKS'] + [i.strip() for i in self.__spc_names]
        self.SPC_NAMES = ''.join([s.ljust(16) for s in self.__spc_names])
        setattr(self, 'VAR-LIST', "".join([i.ljust(16) for i in varkeys]))
        self.variables = PseudoNetCDFVariables(self.__variables, varkeys)
        self.__time_stks()
        self.createDimension('NSTK', self.__nstk_hdr['nstk'])

    def __globalheader(self):
        """
        __readheader reads the header section of the ipr file
        it initializes each header field (see CAMx Users Manual for a list)
        as properties of the ipr class
        """
        ep = self.__endianprefix
        offset = 0

        stop = offset + self.__emiss_hdr_fmt.itemsize // 4
        self.__emiss_hdr = self.__memmap[offset:stop]\
                               .view(self.__emiss_hdr_fmt)
        offset += self.__emiss_hdr.nbytes // 4
        assert((self.__emiss_hdr['SPAD'] == self.__emiss_hdr['EPAD']).all())

        stop = offset + self.__grid_hdr_fmt.itemsize // 4
        self.__grid_hdr = self.__memmap[offset:stop].view(self.__grid_hdr_fmt)
        offset += self.__grid_hdr.nbytes // 4

        assert((self.__grid_hdr['SPAD'] == self.__grid_hdr['EPAD']).all())
        self.NAME = self.__emiss_hdr['name'][0, :, 0].copy().view('S10')[
            0].decode()
        self.NOTE = self.__emiss_hdr['note'][0, :, 0].copy().view('S60')[
            0].decode()
        self.XORIG = self.__grid_hdr['xorg'][0]
        self.YORIG = self.__grid_hdr['yorg'][0]
        self.XCELL = self.__grid_hdr['delx'][0]
        self.YCELL = self.__grid_hdr['dely'][0]
        self.NCOLS = self.__grid_hdr['nx'][0]
        self.NROWS = self.__grid_hdr['ny'][0]
        self.NLAYS = self.__grid_hdr['nz'][0]
        self.PLON = self.__grid_hdr['plon'][0]
        self.PLAT = self.__grid_hdr['plat'][0]
        self.TLAT1 = self.__grid_hdr['tlat1'][0]
        self.TLAT2 = self.__grid_hdr['tlat2'][0]
        self.IUTM = self.__grid_hdr['iutm'][0]
        self.ISTAG = self.__grid_hdr['istag'][0]
        self.CPROJ = self.__grid_hdr['iproj'][0]
        stop = offset + self.__cell_hdr_fmt.itemsize // 4
        self.__cell_hdr = self.__memmap[offset:stop].view(self.__cell_hdr_fmt)
        offset += self.__cell_hdr.nbytes // 4 + 1
        assert((self.__cell_hdr['SPAD'] == self.__cell_hdr['EPAD']).all())

        nspec = self.__emiss_hdr['nspec'][0]
        self.ITZON = self.__emiss_hdr['itzon'][0]

        stop = offset + self.__spc_hdr_fmt.itemsize // 4 * nspec
        self.__spc_hdr = self.__memmap[offset:stop]\
                             .view(ep + 'S1').reshape(nspec, 10, 4)
        offset += self.__spc_hdr.nbytes // 4 + 1

        spc_names = [np.char.strip(spc[:, 0].copy().view('S10'))[
            0] for spc in self.__spc_hdr]
        spc_names = [spc.decode() if hasattr(spc, 'decode')
                     else spc for spc in spc_names]
        self.__spc_names = spc_names
        stop = offset + self.__nstk_hdr_fmt.itemsize // 4
        self.__nstk_hdr = self.__memmap[offset:stop].view(self.__nstk_hdr_fmt)
        offset += self.__nstk_hdr.nbytes // 4 + 1
        assert((self.__nstk_hdr['SPAD'] == self.__nstk_hdr['EPAD']).all())

        nstk = self.__nstk_hdr['nstk'][0]
        self.createDimension('NSTK', nstk)

        stop = offset + self.__stk_prop_fmt.itemsize // 4 * nstk
        self.__stk_props = self.__memmap[offset:stop].reshape(
            nstk, self.__stk_prop_fmt.itemsize // 4).view(self.__stk_prop_fmt)
        offset += self.__stk_props.nbytes // 4 + 1

        self.__data_start = offset

        self.SDATE = self.__emiss_hdr['ibdate']
        self.STIME = self.__emiss_hdr['btime']
        self.EDATE = self.__emiss_hdr['iedate']
        self.ETIME = self.__emiss_hdr['etime']

    def __getspcidx(self, spc):
        return self.__spc_names.index(spc)

    def __time_stks(self):
        ep = self.__endianprefix
        nspcs = len(self.__spc_names)
        nstks = len(self.dimensions['NSTK'])
        date_block_size = 6
        stk_block_size = 4
        stk_props_size = 2 + nstks * 5
        emiss_block_size = nspcs * (nstks + 13)
        hour_block_size = (date_block_size + stk_block_size +
                           stk_props_size + emiss_block_size)
        data = self.__memmap[self.__data_start:]
        data = data.reshape(data.size // hour_block_size, hour_block_size)
        ntimes = data.shape[0]
        self.createDimension('TSTEP', ntimes)
        self.createDimension('DATE-TIME', 2)
        start = 0
        end = date_block_size
        date_times = data[:, start:end]
        dates = date_times[:, [1, 3]].view(ep + 'i')
        times = date_times[:, [2, 4]]

        start = end
        end = start + stk_block_size
        nstk_hdr = data[:, start:end].view(ep + 'i')
        if not (nstks == nstk_hdr[:, 2:3]).all():
            raise ValueError("Number of stacks varies with time")
        start = end
        end = start + stk_props_size
        self.__hourly_stk_props = \
            data[:, start:end][:, 1:-1].reshape(ntimes, nstks, 5)

        start = end
        end = start + emiss_block_size
        if not end == data.shape[1]:
            raise ValueError("Incorrect shape")
        self.__emiss_data = data[:, start:].reshape(
            ntimes, nspcs, 13 + nstks)[:, :, 12:-1]
        bdates = dates[:, 0]
        btimes = times[:, 0]
        edates = dates[:, 1]
        etimes = times[:, 1]
        self.NSTEPS = ntimes

        self.createDimension('VAR', len(self.__spc_names) + 3)

        self.variables['TFLAG'] = ConvertCAMxTime(
            bdates, btimes, len(self.dimensions['VAR']))
        self.variables['ETFLAG'] = ConvertCAMxTime(
            edates, etimes, len(self.dimensions['VAR']))
        v = self.variables['NSTKS'] = PseudoNetCDFVariable(
            self, 'NSTKS', 'i', ('TSTEP',), values=array(nstk_hdr[:, 2]))
        v.units = '#'.ljust(16)
        v.long_name = 'NSTKS'.ljust(16)
        v.var_desc = v.long_name

    def __variables(self, k):
        if k in ['TFLAG', 'ETFLAG', 'NSTKS']:
            return self.variables[k]
        elif k in ['XSTK', 'YSTK', 'HSTK', 'DSTK', 'TSTK', 'VSTK']:
            v = PseudoNetCDFVariable(
                self, k, 'f', ('NSTK',), values=self.__stk_props[k].ravel())
            v.units = {'XSTK': 'm', 'YSTK': 'm', 'HSTK': 'm',
                       'DSTK': 'm', 'TSTK': 'K', 'VSTK': 'm/h'}[k]
            v.long_name = k.ljust(16)
            v.var_desc = k.ljust(16)
            return v
        elif k in ['IONE', 'ITWO', 'KCELL', 'FLOW', 'PLMHT']:
            data_type = {'IONE': 'i', 'ITWO': 'i',
                         'KCELL': 'i', 'FLOW': 'f', 'PLMHT': 'f'}[k]
            v = self.createVariable(k, data_type, ('TSTEP', 'NSTK'))
            v.units = {'IONE': '#', 'ITWO': '#', 'KCELL': '#',
                       'FLOW': 'm**3/hr', 'PLMHT': 'm'}[k]
            v.long_name = k.ljust(16)
            v.var_desc = k.ljust(16)
            vals = self.__hourly_stk_props[:, :, [
                'IONE', 'ITWO', 'KCELL', 'FLOW', 'PLMHT'].index(k)]
            v[:] = vals.view('>' + data_type)
            return v
        elif k in self.__spc_names:
            vals = self.__emiss_data[:, self.__getspcidx(k), :]
            v = PseudoNetCDFVariable(self, k, 'f', ('TSTEP', 'NSTK'),
                                     values=vals)
            v.units = 'mole/hr'.ljust(16)
            v.long_name = k.ljust(16)
            v.var_desc = k.ljust(16)
            return v
        else:
            raise KeyError("Unknown key %s" % k)


class TestMemmap(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testPT(self):
        import PseudoNetCDF.testcase
        emissfile = point_source(
            PseudoNetCDF.testcase.camxfiles_paths['point_source'])
        v = emissfile.variables['NO2']
        self.assertTrue((v[:] == np.array(
            [0.00000000e+00, 3.12931000e+02, 1.23599997e+01, 0.00000000e+00,
             5.27999992e+01, 0.00000000e+00, 3.12931000e+02, 1.23599997e+01,
             0.00000000e+00, 5.27999992e+01], dtype='f').reshape(2, 5)).all())

    def testNCF2PT(self):
        import PseudoNetCDF.testcase
        from PseudoNetCDF.pncgen import pncgen
        import os
        inpath = PseudoNetCDF.testcase.camxfiles_paths['point_source']
        outpath = inpath + '.check'
        infile = point_source(inpath)
        pncgen(infile, outpath, format='camxfiles.point_source')
        orig = open(inpath, 'rb').read()
        new = open(outpath, 'rb').read()
        # outfile=point_source(outpath)
        # for k, v0 in infile.variables.items():
        #    v1 = outfile.variables[k]
        #    print(k, (v0[:] == v1[:]).all())
        assert(orig == new)
        os.remove(outpath)


if __name__ == '__main__':
    unittest.main()



__all__ = ['cloud_rain']
__doc__ = """
.. _Memmap
:mod:`Memmap` -- cloud_rain Memmap interface
============================================

.. module:: Memmap
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map for CAMx
              cloud/rain files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

# Distribution packages
import unittest
import struct
from warnings import warn

# Site-Packages
from numpy import zeros, array, memmap
import numpy as np

# This Package modules
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariable
from PseudoNetCDF.sci_var import PseudoNetCDFVariables
from PseudoNetCDF.ArrayTransforms import ConvertCAMxTime

# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1, ), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class cloud_rain(PseudoNetCDFFile):
    """
    cloud_rain provides a PseudoNetCDF interface for CAMx
    cloud_rain files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> cloud_rain_path = 'cloud_rain.bin'
        >>> rows, cols = 65, 83
        >>> cloud_rainfile = cloud_rain(cloud_rain_path, rows, cols)
        >>> cloud_rainfile.variables.keys()
        ['CLOUD', 'RAIN', 'SNOW', 'GRAUPEL', 'COD', 'TFLAG']
        >>> v = cloud_rainfile.variables['CLOUD']
        >>> tflag = cloud_rainfile.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0, 0, :]
        array([2005185,       0])
        >>> tflag[-1, 0, :]
        array([2005185,  240000])
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> cloud_rainfile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    def __init__(self, rf, rows=None, cols=None):
        f = open(rf, 'rb')
        f.seek(0, 2)
        flen = f.tell()
        offset = struct.unpack('>i', open(rf, 'rb').read(4))[0] + 8
        self.__memmap = memmap(rf, dtype='>f', mode='r', offset=offset)
        cldhdrlen = offset - 20
        line1fmt = '>i%dciiii' % cldhdrlen
        ncols, nrows, nlays = struct.unpack(line1fmt,
                                            open(rf, 'rb').read(offset))[-4:-1]
        self.createDimension('COL', ncols)
        self.createDimension('ROW', nrows)
        self.createDimension('LAY', nlays)
        mydt = '>i4,S%d,>i4,>i4,>i4,>i4,>i4,>f4,>i4' % cldhdrlen
        header = np.fromfile(rf, dtype=mydt, count=1)[0]
        self.FILEDESC = ''.join(header[1].decode())
        self.STIME, self.SDATE = header.tolist()[-2:]
        if self.SDATE < 10000:
            self.SDATE += 2000000
        if (((ncols != cols and cols is not None) or
             (rows != rows and rows is not None))):
            warn(('Files says cols = %d, rows = %d, and lays = %d; ' +
                  'you said cols = %d and rows = %d') % (
                ncols, nrows, nlays, cols, rows))

        self.createDimension('DATE-TIME', 2)

        datasize = (flen - offset)
        # Try 5 first (contemporary)
        # Try 3 second (old)
        # end on 5 as a failsafe
        for nvars in [5, 3, 5]:
            timesize = (nvars * nlays * (nrows * ncols + 2) * 4 + 16)
            if (datasize % timesize) == 0:
                break
        else:
            warn(
                'File appears incomplete using 3 (v4.2) or 5 ' +
                'variables (>=v4.3); expected to fail'
            )

        if nvars < 5:
            self.VERSION = '<4.3'
            varkeys = ['CLOUD', 'PRECIP', 'COD', 'TFLAG']
        else:
            self.VERSION = '>=4.3'
            varkeys = ['CLOUD', 'RAIN', 'SNOW', 'GRAUPEL', 'COD', 'TFLAG']

        ntimes = datasize // timesize
        self.createDimension('TSTEP', ntimes)
        self.createDimension('VAR', len(varkeys) - 1)

        self.NVARS = len(self.dimensions['VAR'])
        self.NLAYS = len(self.dimensions['LAY'])
        self.NROWS = len(self.dimensions['ROW'])
        self.NCOLS = len(self.dimensions['COL'])
        self.FTYPE = 1

        self.variables = PseudoNetCDFVariables(self.__var_get, varkeys)

        self.SDATE, self.STIME = self.variables['TFLAG'][0, 0, :]

    def __set_var(self, key, vals_idx):
        times = len(self.dimensions['TSTEP'])
        lays = len(self.dimensions['LAY'])
        rows = len(self.dimensions['ROW'])
        cols = len(self.dimensions['COL'])
        vals = self.__memmap[vals_idx].reshape(times, lays, rows, cols)
        v = PseudoNetCDFVariable(self, key, 'f',
                                 ('TSTEP', 'LAY', 'ROW', 'COL'),
                                 values=vals)
        v.units = {'COD': 'None'}.get(key, 'g/m**3')
        v.long_name = key
        v.var_desc = key
        self.variables[key] = v

    def __var_get(self, key):
        times = len(self.dimensions['TSTEP'])
        rows = len(self.dimensions['ROW'])
        cols = len(self.dimensions['COL'])
        lays = len(self.dimensions['LAY'])
        vars = len(list(self.variables.keys())) - 1
        hour = 1
        date = 2
        cloud = 3
        rain = 4
        snow = 5
        graupel = 6
        cod = 7
        # stagger = 8
        out_idx = zeros(self.__memmap.shape, dtype='b')
        out_idx.reshape(times, lays * vars *
                        (rows * cols + 2) + 4)[:, 1] = hour
        out_idx.reshape(times, lays * vars *
                        (rows * cols + 2) + 4)[:, 2] = date

        dateblock = self.__memmap[out_idx == date].view('>i')
        hourblock = self.__memmap[out_idx == hour]
        nvars = len(self.dimensions['VAR'])
        self.variables['TFLAG'] = ConvertCAMxTime(dateblock, hourblock, nvars)

        newshape1 = (times, lays * vars * (rows * cols + 2) + 4)
        newshape2 = (times, lays, vars, rows * cols + 2)
        newshape3 = (times, lays, vars, rows, cols)
        val_shape = out_idx.reshape(*newshape1)[:, 4:]
        val_shape = val_shape.reshape(*newshape2)[:, :, :, 1:-1]
        val_shape = val_shape.reshape(*newshape3)
        if self.VERSION == '<4.3':
            val_shape[:, :, 0, :, :] = cloud
            val_shape[:, :, 1, :, :] = rain
            val_shape[:, :, 2, :, :] = cod
            self.__set_var('CLOUD', out_idx == cloud)
            self.__set_var('PRECIP', out_idx == rain)
            self.__set_var('COD', out_idx == cod)
        else:
            val_shape[:, :, 0, :, :] = cloud
            val_shape[:, :, 1, :, :] = rain
            val_shape[:, :, 2, :, :] = snow
            val_shape[:, :, 3, :, :] = graupel
            val_shape[:, :, 4, :, :] = cod
            self.__set_var('CLOUD', out_idx == cloud)
            self.__set_var('RAIN', out_idx == rain)
            self.__set_var('SNOW', out_idx == snow)
            self.__set_var('GRAUPEL', out_idx == graupel)
            self.__set_var('COD', out_idx == cod)

        buf = self.__memmap[out_idx == 0].reshape(
            vars * times * lays + times, 2)
        if not (buf[:, 0] == buf[:, 1]).all():
            raise ValueError("Buffer")

        return self.variables[key]


class TestMemmap(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testCR(self):
        import PseudoNetCDF.testcase
        crfile = cloud_rain(
            PseudoNetCDF.testcase.camxfiles_paths['cloud_rain'], 4, 5)
        crfile.variables['TFLAG']
        checkv = array([1.25412483e+01, 1.77024829e+00, 0.00000000e+00,
                        0.00000000e+00, 0.00000000e+00, 1.38041372e+01,
                        1.94885385e+00, 0.00000000e+00, 0.00000000e+00,
                        0.00000000e+00, 1.41415815e+01, 1.67501605e+00,
                        0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
                        1.34467077e+01, 1.99459922e+00, 0.00000000e+00,
                        0.00000000e+00, 0.00000000e+00, 1.25412483e+01,
                        1.77024829e+00, 0.00000000e+00, 0.00000000e+00,
                        0.00000000e+00, 1.38041372e+01, 1.94885385e+00,
                        0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
                        1.41415815e+01, 1.67501605e+00, 0.00000000e+00,
                        0.00000000e+00, 0.00000000e+00, 1.34467077e+01,
                        1.99459922e+00, 0.00000000e+00, 0.00000000e+00,
                        0.00000000e+00, 1.25412483e+01, 1.77024829e+00,
                        0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
                        1.38041372e+01, 1.94885385e+00, 0.00000000e+00,
                        0.00000000e+00, 0.00000000e+00, 1.41415815e+01,
                        1.67501605e+00, 0.00000000e+00, 0.00000000e+00,
                        0.00000000e+00, 1.34467077e+01, 1.99459922e+00,
                        0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
                        1.96655331e+01, 2.05677104e+00, 0.00000000e+00,
                        0.00000000e+00, 0.00000000e+00, 2.14273071e+01,
                        2.09934115e+00, 0.00000000e+00, 0.00000000e+00,
                        0.00000000e+00, 2.21391239e+01, 0.00000000e+00,
                        0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
                        2.26519203e+01, 4.96763992e+00, 0.00000000e+00,
                        0.00000000e+00, 0.00000000e+00, 1.96655331e+01,
                        2.05677104e+00, 0.00000000e+00, 0.00000000e+00,
                        0.00000000e+00, 2.14273071e+01, 2.09934115e+00,
                        0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
                        2.21391239e+01, 0.00000000e+00, 0.00000000e+00,
                        0.00000000e+00, 0.00000000e+00, 2.26519203e+01,
                        4.96763992e+00, 0.00000000e+00, 0.00000000e+00,
                        0.00000000e+00, 1.96655331e+01, 2.05677104e+00,
                        0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
                        2.14273071e+01, 2.09934115e+00, 0.00000000e+00,
                        0.00000000e+00, 0.00000000e+00, 2.21391239e+01,
                        0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
                        0.00000000e+00, 2.26519203e+01, 4.96763992e+00,
                        0.00000000e+00, 0.00000000e+00, 0.00000000e+00],
                       dtype='f').reshape(2, 3, 4, 5)
        self.assertTrue((crfile.variables['COD'] == checkv).all())

    def testNCF2CR(self):
        import PseudoNetCDF.testcase
        from PseudoNetCDF.pncgen import pncgen
        import os
        inpath = PseudoNetCDF.testcase.camxfiles_paths['cloud_rain']
        outpath = inpath + '.check'
        infile = cloud_rain(inpath, 4, 5)
        pncgen(infile, outpath, format='camxfiles.cloud_rain')
        orig = open(inpath, 'rb').read()
        new = open(outpath, 'rb').read()
        assert(orig == new)
        os.remove(outpath)


if __name__ == '__main__':
    unittest.main()



__all__ = ['cartesian', 'sliceit']
__doc__ = """
.. _util
:mod:`util` -- CAMx basic util
==============================

.. module:: util
   :platform: Unix, Windows
   :synopsis: Provides simple utilites for camxfiles
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""


def cartesian(x, y):
    """Iterator for an 'outer' or cartesian join of
    x iterator and y iterator
    """
    for i in x:
        for j in y:
            yield i, j


def sliceit(args):
    """
    If arguments are a slice, return it.  Otherwise,
    create a slice.
    """
    if isinstance(args, slice):
        return args

    try:
        return slice(*args)
    except TypeError:
        return slice(args, args + 1)



import PseudoNetCDF as pnc
from PseudoNetCDF.pncwarn import warn
import os
import io
from datetime import datetime
from collections import OrderedDict
import numpy as np
from ._ioapi import ioapi_base

_prjp = ('GDTYP', 'P_ALP', 'P_BET', 'P_GAM', 'XCENT', 'YCENT')
_grdp = (
    'PRJNAME', 'XORIG', 'YORIG', 'XCELL', 'YCELL', 'NCOLS', 'NROWS', 'NTHIK'
)


class griddesc(ioapi_base):
    """
    griddesc is designed to read IOAPI griddesc files and create
    a file with basic CF metadata. An example format of GRIDDESC
    is show below with two grids on one projection.

' '
'LCC'
  2        33.000        45.000       -97.000       -97.000        40.000
' '
'SE52BENCH'
'LCC'    792000.000  -1080000.000     12000.000     12000.000 100  80   1
'36US3'
'LCC'  -2952000.000  -2772000.000     36000.000     36000.000 172  148   1
' '
"""
    def __init__(
        self, path, GDNAM=None, VGLVLS=(1., 0.), VGTOP=5000., FTYPE=1, VGTYP=7,
        SDATE=-635, STIME=0, TSTEP=0, var_kwds=None, nsteps=1, withcf=True,
        **prop_kw
    ):
        """
        Arguments
        ---------
        path : str or file-like
            If path is has a read method, it will be used directly.
            If path is a path to a file on disk, it will be read.
            If none of these, it will try to treat the path as content.
        GDNAM : str
            Name the grid to be used. If not provided, the first will be used
        VGLVLS : tuple
            Iterable of layer edge (k+1) values for the vertical grid.
        VGTOP : float
            Top of vertical grid.
        VGTYP : int
            Determines the units of VGLVLS and VGTOP. (7: WRF sigma-p,
            6: meters asl, for more details see IOAPI documentation)
        SDATE : int
            Starting julian date (YYYYJJJ) or -635
        STIME : int
            Starting time as HHMMSS
        TSTEP : int
            IOAPI time step in HHMMSS
        FTYPE : int
            1 for gridded; 2 for boundary
        var_kwds: tuple, list, or dict
            See setdefvars; Defaults to {'DUMMY': {'units': 'unknown'}}
        nsteps : int
            Number of time steps to use. Should be coupled with SDATE, STIME,
            and TSTEP. nsteps > 1 is not compatible with time independent
            (TSTEP=0) or SDATE=-635
        withcf : bool
            If true, then CF compatible variables that describe dimensions
            and time are added.
        prop_kw: mappable
            Can provide additional IOAPI (or other) properties for the file.
        """
        now = datetime.now()
        if hasattr(path, 'read'):
            gdf = path
            path = '<inline>'
        elif os.path.exists(path):
            gdf = open(path, 'r')
        else:
            gdf = io.StringIO(path)
            path = '<inline>'
        gdlines = gdf.read().strip().split('\n')
        gdlines = [line.strip() for line in gdlines]
        assert(gdlines[0] == "' '")
        assert(gdlines[-1] == "' '")
        i = 0
        prj = {}
        grd = OrderedDict()
        blanks = []
        while i < len(gdlines):
            line = gdlines[i]
            if line.strip() == "' '":
                blanks.append(i)
            else:
                i += 1
                parts = [eval(p) for p in gdlines[i].split()]
                key = eval(line)
                if len(blanks) == 1:
                    prj[key] = dict(zip(_prjp, parts))
                elif len(blanks) == 2:
                    grd[key] = dict(zip(_grdp, parts))
                else:
                    pass
            i += 1
        self._prj = prj
        self._grd = grd
        self.FTYPE = FTYPE
        self.VGTYP = VGTYP
        self.GDNAM = GDNAM
        self.VGTOP = np.float32(VGTOP)
        self.VGLVLS = np.array(VGLVLS, dtype='f')
        self.NLAYS = len(VGLVLS) - 1
        self.UPNAM = 'GRIDDESC'
        self.EXEC_ID = 'GRIDDESC'
        self.FILEDESC = 'GRIDDESC'
        self.HISTORY = 'made from ' + path
        self.CDATE = int(now.strftime('%Y%j'))
        self.CTIME = int(now.strftime('%H%M%S'))
        self.WDATE = self.CDATE
        self.WTIME = self.CTIME
        self.SDATE = SDATE
        self.STIME = STIME
        self.TSTEP = TSTEP
        self._synthvars = None
        # user supplied properties should overwrite
        # defaults above
        for pk, pv in prop_kw.items():
            self.setncattr(pk, pv)

        self.setdefvars(var_kwds)
        self.addtime(nsteps)
        self.setgrid(withcf=withcf)
        self.updatemeta()
        self.getVarlist()

    def setdefvars(self, var_kwds):
        """
        Arguments
        ---------
        var_kwds: list, tuple, or dictionary
            Variables with units or just a list of variables
        """

        if var_kwds is None:
            var_kwds = {'DUMMY': {'units': 'unknown'}}
        elif isinstance(var_kwds, (tuple, list)):
            var_kwds = {k: 'unknown' for k in var_kwds}
        elif hasattr(var_kwds, 'items'):
            pass
        else:
            raise TypeError('var_kwds should be a list, tuple, or dictionary')

        self._var_kwds = {}
        for k, v in var_kwds.items():
            if isinstance(v, str):
                self._var_kwds[k] = {'units': v.ljust(16)}
            elif isinstance(v, dict):
                self._var_kwds[k] = {_k: _v for _k, _v in v.items()}
            else:
                raise TypeError(
                    'var_kwds should have either str values (units) or a dict'
                    + ' of all properties'
                )

    def setgrid(self, key=None, withcf=True):
        """
        Remakes the file to use the grid specified by key

        Arguments
        ---------
        key : str
            GDNAM to set the grid of the file
        withcf : bool
            Passed to adddims

        Returns
        -------
        None
        """
        if key is None:
            if self.GDNAM is None:
                self.GDNAM = list(self._grd)[0]

            key = self.GDNAM.strip()
        grd = self._grd[key]
        prj = self._prj[grd['PRJNAME'].strip()]
        self.setncatts(prj)
        self.setncatts(grd)
        self.adddims(withcf=withcf)

    def addtime(self, nsteps=1):
        """
        Adds TFLAG variable

        Arguments
        ---------
        nsteps : int
            Number of time steps to add

        Returns
        -------
        None
        """
        self.createDimension('TSTEP', nsteps).setunlimited(True)
        self.createDimension('DATE-TIME', 2)
        self.NVARS = len(self._var_kwds)
        self.createDimension('VAR', self.NVARS)
        try:
            self.updatetflag(overwrite=True)
        except Exception as e:
            warn(str(e))
            tflag = self.createVariable(
                'TFLAG', 'i',
                ('TSTEP', 'VAR', 'DATE-TIME'),
                units='<YYYYJJJ,HHMMSS>',
                long_name='TFLAG'.ljust(16),
                var_desc='TFLAG'.ljust(80)
            )
            tflag[:, :, 0] = self.SDATE
            tflag[:, :, 1] = 0

    def adddims(self, withcf=True):
        """
        Add spatial dimensions

        Arguments
        ---------
        withcf : bool
            If true, add Climate and Forecasting convention variables

        Returns
        -------
        None
        """
        for k in 'LAY ROW COL PERIM'.split():
            if k in self.dimensions:
                del self.dimensions[k]
        self.createDimension('LAY', self.NLAYS)
        if self.FTYPE == 1:
            self.createDimension('ROW', self.NROWS)
            self.createDimension('COL', self.NCOLS)
            dims = ('TSTEP', 'LAY', 'ROW', 'COL')
        elif self.FTYPE == 2:
            perim = (self.NCOLS * 2 + self.NROWS * 2 + self.NTHIK * 4)
            self.createDimension('PERIM', perim * self.NTHIK)
            dims = ('TSTEP', 'LAY', 'PERIM')
        else:
            raise ValueError(
                'Only supports FTYPE 1 or 2; received ' +
                str(self.FTYPE)
            )

        for vark, varkw in self._var_kwds.items():
            self.createVariable(
                vark, 'f', dims,
                long_name=vark.ljust(16),
                var_desc=vark.ljust(80),
                **varkw
            )

        if self._synthvars is None:
            oldkeys = set(self.variables)
        else:
            for key in self._synthvars:
                if key in self.variables:
                    del self.variables[key]
        if withcf:
            pnc.conventions.ioapi.add_cf_from_ioapi(self)
        if self._synthvars is None:
            self._synthvars = set(self.variables).difference(oldkeys)


if __name__ == '__main__':
    f = griddesc('GRIDDESC', '36US3')



__all__ = ['getreader', 'pncopen', 'pncmfopen']

import unittest
import os
from warnings import warn
from PseudoNetCDF.netcdf import NetCDFFile

_readers = [('Dataset', NetCDFFile)]


def testreader(reader, *args, **kwds):
    try:
        reader(*args, **kwds)
        return True
    except Exception:
        return False


def getreader(*args, **kwds):
    """
    Parameters
    ----------
    *args :
        arguments for opening file
    **kwds :
        keywords for file opener and optional format
    format : string
        name of reader (optional)

    Returns
    -------
    reader : class
    """
    global _readers
    format = kwds.pop('format', None)
    if not os.path.isfile(args[0]):
        warn(('The first argument (%s) does not exist as a file.  ' +
              'First arguments are usually paths') % (args[0],))

    if format is None:
        _myreaders = _readers
        # Try to give preferential treatment based on suffix
        try:
            if len(args) > 0:
                ext = os.path.splitext(args[0])[1][1:]
            else:
                ext = os.path.splitext(kwds['path'])[1][1:]
            rdict = getreaderdict()
            if ext in rdict:
                _myreaders.insert(0, (ext, rdict[ext]))
        except Exception:
            pass
    else:
        _myreaders = [(k, v) for k, v in _readers if format == k]

    for rn, reader in _myreaders:
        if hasattr(reader, 'isMine'):
            checker = reader.isMine
        else:
            def checker(*args, **kwds):
                try:
                    testreader(reader, *args, **kwds)
                    return True
                except Exception:
                    return False
        if checker(*args, **kwds):
            return reader
    else:
        raise TypeError(('No reader could open a file with these ' +
                         'arguments %s %s') % (args, kwds))


def registerreader(name, reader):
    global _readers, pncopen
    if name not in [k for k, v in _readers]:
        _readers.insert(0, (name, reader))
        if pncopen.__doc__ is not None:
            newdoc = '\n        - '.join([pncopen.__doc__, name])
            pncopen.__doc__ = newdoc
        return True
    else:
        return False


def pncmfopen(*args, stackdim=None, **kwds):
    """Open any PNC supported format using pncopen on all files
    passed as teh first argument of pncmfopen. See pncopen

    Parameters
    ----------
    args : arguments
        args[0] must be a list of paths, other arguments are format specific
    stackdim : str
        dimension upon which to stack files
    kwds : dict
        see pncopen for more details

    Returns
    -------
    pfile : PseudoNetCDF
    """
    paths = args[0]
    files = [pncopen(path, *args[1:], **kwds) for path in paths]
    file1 = files[0]
    return file1.stack(files[1:], stackdim=stackdim)


def pncopen(*args, **kwds):
    """Open any PNC supported format using args and kwds, which
    are format specific. format is not passed to the reader

    Parameters
    ----------
    *args :
        arguments for opening file
    **kwds :
        keywords for reader (see format)
    format : string
        name of reader (not passed to reader), default auto-detect
        see Format Options for formats
    addcf : boolean
        to add CF conventions (not passed to reader; default: False)
    diskless : boolean
        to add CF conventions (not passed to reader; default: False)
    help : boolean
        without format, returns help of pncopen and with format keyword,
        returns help of that class. See the __init__ interface for help
        with keywords that are not in the class __doc__.

    Returns
    -------
    pfile : PseudoNetCDF

    Notes
    -----
    Format Options:
        - add your own by subclassing PseudoNetCDFFile
        - for a full list, use pncopen(help=True)
    """
    formathelp = kwds.pop('help', False)
    format = kwds.pop('format', None)
    addcf = kwds.pop('addcf', False)
    diskless = kwds.pop('diskless', False)
    if format is None:
        if formathelp:
            return help(pncopen)
        reader = getreader(*args, format=format, **kwds)
    else:
        reader = getreaderdict()[format]
        if formathelp:
            return help(reader)

    outfile = reader(*args, **kwds)
    if addcf:
        try:
            from .conventions.ioapi import add_cf_from_ioapi
            from .core._wrapnc import WrapPNC
            from .core._functions import getvarpnc
            if not diskless:
                outfile = WrapPNC(outfile)
            else:
                outfile = getvarpnc(outfile, None)
            add_cf_from_ioapi(outfile)
        except Exception as e:
            warn(str(e))
            pass
    return outfile


anyfile = pncopen


def getreaderdict():
    return dict(_readers)


class TestPNCOPEN(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testPNCOPENFMT(self):
        import PseudoNetCDF.testcase
        for format, path in PseudoNetCDF.testcase.self_described_paths.items():
            print('Test open with ', format, path)
            pncopen(path, format=format)

    def testPNCOPENNOFMT(self):
        import PseudoNetCDF.testcase
        for format, path in PseudoNetCDF.testcase.self_described_paths.items():
            print('Test open unspecified ', path)
            pncopen(path)
